
/*
 * tr2dp.c
 *
 * Tomb Raider II dump routines for TRueView
 *
 *** CURRENTLY LACKING:
 ***     Better error control (specifically, any malloc() or fread() error
 ***                           just unceremoniously exits the program)
 ***     Correct handling for unknowns (obviously, I can't correctly
 ***                           handle what I don't understand)
 */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <memory.h>
#include <string.h>
#include <stdlib.h>

#include "config.h"

#ifdef INCLUDE_TR4_SUPPORT
// ZLIB #defines and #include for TR4 compression support
#ifndef LINUX
#define _WINDOWS 1
#define ZLIB_DLL 1
#endif
#include "zlib.h"
#endif // INCLUDE_TR4_SUPPORT

#include "tr2port.h"
#include "tr2.h"

/*
 * definitions, declarations, prototypes, etc.
 */
// prototype of hex dump routine
void HexDump(unsigned char *buffer, int len, int offset, char *prefix, FILE *fp);

/*
 * HexDump() does what its name implies - a 16-byte-wide dump, with offsets
 * starting at the specified value.
 */
void HexDump(unsigned char *buffer, int len, int offset, char *prefix, FILE *fp)
{
   int i;

   if (len != 0) {   // don't display anything if it's zero-length
      do {
         fprintf(fp, "%s%04x   ", prefix, offset);
         offset += 16;
         for (i = 0; i < 16 && len--; ++i) {
            fprintf(fp, "%02x ", *buffer++);
            }
         fprintf(fp, "\n");
         } while (len > 0);
      }
}

/*
 * DumpTR2level(TR2_Level *Level, char *FileName)
 *
 * Creates <FileName> and dumps readable ASCII level data to it.
 * Print an easy-to-find marker (e.g. "-=-=-=-=-=" at every heading.
 */
int DumpTR2level(TR2_Level *Level, char *FileName)
{
   FILE *fp;
   int  ReturnValue,
        i,
        j,
        k,
        l;

   if ((fp = fopen(FileName, WRITETEXT)) != NULL) {
      fprintf(fp, "Dump of data for '%s'\n\n", Level->FileName);
      fprintf(fp, "Version = %ld (0x%08lx)\n", Level->Version, Level->Version);
      fprintf(fp, "8-bit palette:\n");
      for (i = 0; i < 256; ++i) {
         fprintf(fp, "   Palette8[%3d]: Red %3d  Green %3d  Blue %3d\n", i, Level->Palette8[i].Red, Level->Palette8[i].Green, Level->Palette8[i].Blue);
         }
      fprintf(fp, "16-bit palette: -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n");
      HexDump((bitu8 *)Level->Palette16, sizeof(Level->Palette16), 0, "", fp);
// skip textiles - hex dumps of graphic bitmaps aren't much fun to look at...
fprintf(fp, "(Skipped %d 8-bit Textiles) -=-=-=-=-=-=-=-=-=-=-=-\n", Level->NumTextiles);
fprintf(fp, "(Skipped %d 16-bit Textiles) -=-=-=-=-=-=-=-=-=-=-=\n", Level->NumTextiles);
      fprintf(fp, "UnknownT = %ld (0x%08lx)\n", Level->UnknownT, Level->UnknownT);
      fprintf(fp, "NumRooms: %d -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n", Level->NumRooms);
      for (i = 0; i < Level->NumRooms; ++i) {
         fprintf(fp, "Room %d:  ---===---\n", i);
         fprintf(fp, "   Info: X %d  Z %d  yBottom %d  yTop %d\n", Level->Rooms[i].info.x, Level->Rooms[i].info.z, Level->Rooms[i].info.yBottom, Level->Rooms[i].info.yTop);
         fprintf(fp, "   NumVertices: %d\n", Level->Rooms[i].RoomData.NumVertices);
         for (j = 0; j < Level->Rooms[i].RoomData.NumVertices; ++j) {
            fprintf(fp, "      Vertex[%3d]: (%5d, %5d, %5d)  Lighting1 %d  Attributes 0x%04x  Lighting2 %d\n", j,
                   Level->Rooms[i].RoomData.Vertices[j].Vertex.x,
                   Level->Rooms[i].RoomData.Vertices[j].Vertex.y,
                   Level->Rooms[i].RoomData.Vertices[j].Vertex.z,
                   Level->Rooms[i].RoomData.Vertices[j].Lighting1,
                   Level->Rooms[i].RoomData.Vertices[j].Attributes,
                   Level->Rooms[i].RoomData.Vertices[j].Lighting2);
            }
         fprintf(fp, "   NumRectangles: %d\n", Level->Rooms[i].RoomData.NumRectangles);
         for (j = 0; j < Level->Rooms[i].RoomData.NumRectangles; ++j) {
            fprintf(fp, "      Rectangles[%3d]: Vertices (%3d %3d %3d %3d)  Texture %d\n", j,
                   Level->Rooms[i].RoomData.Rectangles[j].Vertices[0],
                   Level->Rooms[i].RoomData.Rectangles[j].Vertices[1],
                   Level->Rooms[i].RoomData.Rectangles[j].Vertices[2],
                   Level->Rooms[i].RoomData.Rectangles[j].Vertices[3],
                   Level->Rooms[i].RoomData.Rectangles[j].Texture);

            }
         fprintf(fp, "   NumTriangles: %d\n", Level->Rooms[i].RoomData.NumTriangles);
         for (j = 0; j < Level->Rooms[i].RoomData.NumTriangles; ++j) {
            fprintf(fp, "      Triangles[%3d]: Vertices (%3d %3d %3d %3d)  Texture %d\n", j,
                   Level->Rooms[i].RoomData.Triangles[j].Vertices[0],
                   Level->Rooms[i].RoomData.Triangles[j].Vertices[1],
                   Level->Rooms[i].RoomData.Triangles[j].Vertices[2],
                   Level->Rooms[i].RoomData.Triangles[j].Texture);

            }
         fprintf(fp, "   NumSprites: %d\n", Level->Rooms[i].RoomData.NumSprites);
         for (j = 0; j < Level->Rooms[i].RoomData.NumSprites; ++j) {
            fprintf(fp, "      Sprites[%3d]: Vertex %3d  Texture %d\n", j,
                   Level->Rooms[i].RoomData.Sprites[j].Vertex,
                   Level->Rooms[i].RoomData.Sprites[j].Texture);

            }
         fprintf(fp, "   NumPortals: %d\n", Level->Rooms[i].NumPortals);
         for (j = 0; j < Level->Rooms[i].NumPortals; ++j) {
            fprintf(fp, "      Portals[%2d]: Adjoining Room %2d  Normal (%3d, %3d, %3d)\n", j,
                   Level->Rooms[i].Portals[j].AdjoiningRoom,
                   Level->Rooms[i].Portals[j].Normal.x,
                   Level->Rooms[i].Portals[j].Normal.y,
                   Level->Rooms[i].Portals[j].Normal.z);
            fprintf(fp, "                 Vertices (%5d, %5d, %5d) (%5d, %5d, %5d) (%5d, %5d, %5d) (%5d, %5d, %5d)\n",
                   Level->Rooms[i].Portals[j].Vertices[0].x,
                   Level->Rooms[i].Portals[j].Vertices[0].y,
                   Level->Rooms[i].Portals[j].Vertices[0].z,
                   Level->Rooms[i].Portals[j].Vertices[1].x,
                   Level->Rooms[i].Portals[j].Vertices[1].y,
                   Level->Rooms[i].Portals[j].Vertices[1].z,
                   Level->Rooms[i].Portals[j].Vertices[2].x,
                   Level->Rooms[i].Portals[j].Vertices[2].y,
                   Level->Rooms[i].Portals[j].Vertices[2].z,
                   Level->Rooms[i].Portals[j].Vertices[3].x,
                   Level->Rooms[i].Portals[j].Vertices[3].y,
                   Level->Rooms[i].Portals[j].Vertices[3].z);
            }
         fprintf(fp, "   NumZsectors: %d  NumXsectors: %d\n", Level->Rooms[i].NumZsectors, Level->Rooms[i].NumXsectors);
         for (j = 0; j < (Level->Rooms[i].NumZsectors * Level->Rooms[i].NumXsectors); ++j) {
            fprintf(fp, "      Sector[%3d]: FDindex %5d  BoxIndex %d  RoomBelow %3d  Floor %4d  RoomAbove %3d  Ceiling %4d\n", j,
                    Level->Rooms[i].SectorList[j].FDindex,
                    Level->Rooms[i].SectorList[j].BoxIndex,
                    Level->Rooms[i].SectorList[j].RoomBelow,
                    Level->Rooms[i].SectorList[j].Floor,
                    Level->Rooms[i].SectorList[j].RoomAbove,
                    Level->Rooms[i].SectorList[j].Ceiling);
            }
         fprintf(fp, "   SectorData (6):\n");
         HexDump((bitu8 *)&Level->Rooms[i].Intensity1, 6, 0, "      ", fp);
         fprintf(fp, "   NumLights: %d\n", Level->Rooms[i].NumLights);
         for (j = 0; j < Level->Rooms[i].NumLights; ++j) {
            fprintf(fp, "      Light %2d: pos (%5d, %5d, %5d) Intensity1 %d  Intensity2 %d  Fade1 %d  Fade2 %d\n", j,
                    Level->Rooms[i].Lights[j].x, Level->Rooms[i].Lights[j].y, Level->Rooms[i].Lights[j].z,
                    Level->Rooms[i].Lights[j].Intensity1, Level->Rooms[i].Lights[j].Intensity2,
                    Level->Rooms[i].Lights[j].Fade1, Level->Rooms[i].Lights[j].Fade2);
            }
         fprintf(fp, "   NumStaticMeshes: %d\n", Level->Rooms[i].NumStaticMeshes);
         for (j = 0; j < Level->Rooms[i].NumStaticMeshes; ++j) {
            fprintf(fp, "      StaticMesh %2d: pos (%6d, %6d, %6d)  Rotation 0x%04x  Unknown2 %d  Unknown3 %d  ObjectID %d\n", j,
                    Level->Rooms[i].StaticMeshes[j].x, Level->Rooms[i].StaticMeshes[j].y, Level->Rooms[i].StaticMeshes[j].z,
                    Level->Rooms[i].StaticMeshes[j].Rotation, Level->Rooms[i].StaticMeshes[j].Intensity1, Level->Rooms[i].StaticMeshes[j].Intensity2,
                    Level->Rooms[i].StaticMeshes[j].ObjectID);
            }
//         HexDump(Level->Rooms[i].LightData, Level->Rooms[i].NumLightData * 20, 0, "      ", fp);
         fprintf(fp, "   AlternateRoom: %d  Flags: %d\n", Level->Rooms[i].AlternateRoom, Level->Rooms[i].Flags);
         if (Level->EngineVersion >= TombRaider_3) {
            fprintf(fp, "   RoomLightColour: R %d G %d B %d\n", Level->Rooms[i].RoomLightColour.Red, Level->Rooms[i].RoomLightColour.Green, Level->Rooms[i].RoomLightColour.Blue);
            }
         }
#ifdef NEVER_COMPILE_THIS
      fprintf(fp, "NumFloorData: %d -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n", Level->NumFloorData);
      for (i = 0; i < (int)Level->NumFloorData; ++i) {
         fprintf(fp, "   FloorData[%4d] = 0x%04x\n", i, Level->FloorData[i]);
         }
#endif
      // FloorData parser
      fprintf(fp, "Parsed FloorData: -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n");
      fprintf(fp, "   FloorData[   0] = 0x%04x (ignored)\n", Level->FloorData[0]);
      for (i = 1; i < (int)Level->NumFloorData; ) {
         bitu8 Function, SubFunction, More;

         Function = Level->FloorData[i] & 0xff;
         SubFunction = (Level->FloorData[i] & 0x7fff) >> 8;
         More = ((Level->FloorData[i] & 0x8000) == 0);
         fprintf(fp, "   FloorData[%4d]: 0x%04x ", i, Level->FloorData[i]);
         ++i;
         switch (Function) {
            case 0x01 : // door
               fprintf(fp, "0x%04x  Door to room %d (SubFunction 0x%02x)\n", Level->FloorData[i], Level->FloorData[i], SubFunction);
               ++i;     // skip room number
               break;
            case 0x02 : // floor slant
               fprintf(fp, "0x%04x  Floor Slant   UD %4d  LR %4d (SubFunction 0x%02x)\n", Level->FloorData[i], (bit8)((Level->FloorData[i] >> 8) & 0xff), (bit8)(Level->FloorData[i] & 0xff), SubFunction);
               ++i;     // skip slant value
               break;
            case 0x03 : // ceiling slant
               fprintf(fp, "0x%04x  Ceiling Slant UD %4d  LR %4d (SubFunction 0x%02x)\n", Level->FloorData[i], (bit8)((Level->FloorData[i] >> 8) & 0xff), (bit8)(Level->FloorData[i] & 0xff), SubFunction);
               ++i;     // skip slant value
               break;
            case 0x04 : // fdlist action
               fprintf(fp, "0x%04x  ", Level->FloorData[i]);
               switch (SubFunction) {
                  case 0x00 :
                     fprintf(fp, "Run FDlist (ACTIVATE)");
                     break;
                  case 0x01 :
                     fprintf(fp, "If Lara on Ground, Run FDlist (ACTIVATE)");
                     break;
                  case 0x02 :
                     fprintf(fp, "Activate/Deactivate FDlist+1 if item %d active/inactive", Level->FloorData[i + 1] & 0x03ff);
                     break;
                  case 0x03 :
                     fprintf(fp, "Activate FDlist+1 if item %d active", Level->FloorData[i + 1] & 0x03ff);
                     break;
                  case 0x04 :
                     fprintf(fp, "Activate FDlist+1 if item %d is picked up", Level->FloorData[i + 1] & 0x03ff);
                     break;
                  case 0x06 :
                     fprintf(fp, "If Lara on Ground, Run FDlist (DEactivate)");
                     break;
                  case 0x09 :
                     fprintf(fp, "Run FDlist (DEactivate)");
                     break;
                  default :
                     fprintf(fp, "[unknown FDlist SubFunction 0x%02x]", SubFunction);
                     break;
                  }
               fprintf(fp, " (unknown1 = %d)\n", Level->FloorData[i]);
               ++i;     // skip unknown1
               j = 0;
               do {
                  bitu8 command;
                  fprintf(fp, "       FDlist[%3d]: 0x%04x ", j++, Level->FloorData[i]);
                  command = (Level->FloorData[i] >> 10) & 0x1f;
                  switch (command) {
                     case 0x00 :
                        fprintf(fp, "Activate/Deactivate item %d\n", Level->FloorData[i] & 0x03ff);
                        break;
                     case 0x01 :
                        fprintf(fp, "Switch to camera [%d]\n", Level->FloorData[i] & 0x03ff);
                        break;
                     case 0x02 :
                        k = Level->FloorData[i] & 0x03ff;
                        fprintf(fp, "Camera Delay %d sec, %s, unk bit %d\n", k & 0xff, (k & 0x100) ? "ONCE" : "Every Time", (k & 0x200) != 0);
                        break;
                     case 0x06 :
                        fprintf(fp, "Look at item %d\n", Level->FloorData[i] & 0x03ff);
                        break;
                     case 0x07 :
                        fprintf(fp, "End Level\n");
                        break;
                     default :
                        fprintf(fp, "Unknown command 0x%02x, data 0x%04x\n", command, Level->FloorData[i] & 0x03ff); 
                        break;
                     }
                  } while ((Level->FloorData[i++] & 0x8000) == 0);
               break;
            case 0x05 : // kills Lara
               fprintf(fp, "        Kills Lara - SubFunction 0x%02x\n", SubFunction);
               break;
            case 0x06 : // used for walls that are climbable
               fprintf(fp, "        Climbable Wall%s:", (SubFunction == 1 || SubFunction == 2 || SubFunction == 4 || SubFunction == 8) ? "" : "s");
               if (SubFunction & 0x02)
                  fprintf(fp, " +X");
               if (SubFunction & 0x01)
                  fprintf(fp, " +Z");
               if (SubFunction & 0x08)
                  fprintf(fp, " -X");
               if (SubFunction & 0x04)
                  fprintf(fp, " -Z");
               fprintf(fp, "\n");
               break;
            }
         if (!More)
            fprintf(fp, "   --- end ---\n");
         }
      fprintf(fp, "NumMeshes: %d -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n", Level->NumMeshes);
      for (i = 0; i < Level->NumMeshes; ++i) {
         fprintf(fp, "Meshes[%3d]: Centre (%5d, %5d, %5d)\n", i, Level->Meshes[i].Centre.x, Level->Meshes[i].Centre.y, Level->Meshes[i].Centre.z);
         fprintf(fp, "      CollisionSize: 0x08lx\n", Level->Meshes[i].CollisionSize);
         fprintf(fp, "      NumVertices: %d\n", Level->Meshes[i].NumVertices);
         for (j = 0; j < Level->Meshes[i].NumVertices; ++j) {
            fprintf(fp, "         Vertex[%3d]: (%5d, %5d, %5d)\n", j,
                   Level->Meshes[i].Vertices[j].x,
                   Level->Meshes[i].Vertices[j].y,
                   Level->Meshes[i].Vertices[j].z);
            }
         fprintf(fp, "      NumNormals: %d\n", Level->Meshes[i].NumNormals * ((Level->Meshes[i].MeshLights != NULL) ? -1 : 1));
         for (j = 0; j < Level->Meshes[i].NumNormals; ++j) {
            if (Level->Meshes[i].MeshLights != NULL) {
               fprintf(fp, "         MeshLights[%3d]: %5d\n", j, Level->Meshes[i].MeshLights[j]);
               }
            else {
               fprintf(fp, "         Normal[%3d]: (%5d, %5d, %5d)\n", j,
                       Level->Meshes[i].Normals[j].x,
                       Level->Meshes[i].Normals[j].y,
                       Level->Meshes[i].Normals[j].z);
               }
            }
         fprintf(fp, "   NumTexturedRectangles: %d\n", Level->Meshes[i].NumTexturedRectangles);
         for (j = 0; j < Level->Meshes[i].NumTexturedRectangles; ++j) {
            fprintf(fp, "         TexturedRectangles[%3d]: Vertices (%3d %3d %3d %3d) Texture %d\n", j,
                   Level->Meshes[i].TexturedRectangles[j].Vertices[0],
                   Level->Meshes[i].TexturedRectangles[j].Vertices[1],
                   Level->Meshes[i].TexturedRectangles[j].Vertices[2],
                   Level->Meshes[i].TexturedRectangles[j].Vertices[3],
                   Level->Meshes[i].TexturedRectangles[j].Texture);

            }
         fprintf(fp, "   NumTexturedTriangles: %d\n", Level->Meshes[i].NumTexturedTriangles);
         for (j = 0; j < Level->Meshes[i].NumTexturedTriangles; ++j) {
            fprintf(fp, "         TexturedTriangles[%3d]: Vertices (%3d %3d %3d) Texture %d\n", j,
                   Level->Meshes[i].TexturedTriangles[j].Vertices[0],
                   Level->Meshes[i].TexturedTriangles[j].Vertices[1],
                   Level->Meshes[i].TexturedTriangles[j].Vertices[2],
                   Level->Meshes[i].TexturedTriangles[j].Texture);

            }
         fprintf(fp, "   NumColouredRectangles: %d\n", Level->Meshes[i].NumColouredRectangles);
         for (j = 0; j < Level->Meshes[i].NumColouredRectangles; ++j) {
            fprintf(fp, "         ColouredRectangles[%3d]: Vertices (%3d %3d %3d %3d) Colour 0x%02x (=%d?)\n", j,
                   Level->Meshes[i].ColouredRectangles[j].Vertices[0],
                   Level->Meshes[i].ColouredRectangles[j].Vertices[1],
                   Level->Meshes[i].ColouredRectangles[j].Vertices[2],
                   Level->Meshes[i].ColouredRectangles[j].Vertices[3],
                   Level->Meshes[i].ColouredRectangles[j].Texture,
                   Level->Meshes[i].ColouredRectangles[j].Texture & 0xff);

            }
         fprintf(fp, "   NumColouredTriangles: %d\n", Level->Meshes[i].NumColouredTriangles);
         for (j = 0; j < Level->Meshes[i].NumColouredTriangles; ++j) {
            fprintf(fp, "         ColouredTriangles[%3d]: Vertices (%3d %3d %3d) Colour 0x%02x (=%d?)\n", j,
                   Level->Meshes[i].ColouredTriangles[j].Vertices[0],
                   Level->Meshes[i].ColouredTriangles[j].Vertices[1],
                   Level->Meshes[i].ColouredTriangles[j].Vertices[2],
                   Level->Meshes[i].ColouredTriangles[j].Texture,
                   Level->Meshes[i].ColouredTriangles[j].Texture & 0xff);

            }
         }
      fprintf(fp, "NumAnimations: %d\n", Level->NumAnimations);
      for (i = 0; i < (int)Level->NumAnimations; ++i) {
         fprintf(fp, "Animations[%3d]: FrameOffset %d  (FrameIndex %d)  Unknown1 %d  FrameSize %d\n", i, 
                 Level->Animations[i].FrameOffset, Level->Animations[i].FrameOffset / 2, Level->Animations[i].Unknown1, Level->Animations[i].FrameSize);
         fprintf(fp, "          Unknown2 (10):\n");
         HexDump((bitu8 *)&Level->Animations[i].Unknown1, 8, 0, "          ", fp);
         fprintf(fp, "          FrameStart %d  FrameEnd %d\n", Level->Animations[i].FrameStart, Level->Animations[i].FrameEnd);
         fprintf(fp, "          AnimationChain Anim:Frame %d:%d\n",
                 Level->Animations[i].NextAnimation,
                 Level->Animations[i].NextFrame);
         fprintf(fp, "          StartingStateChange %d  NumStateChanges %d\n",
                 Level->Animations[i].StateChangeOffset,
                 Level->Animations[i].NumStateChanges);
         fprintf(fp, "          NumAnimCommands %d  AnimCommands %d\n",
                 Level->Animations[i].NumAnimCommands,
                 Level->Animations[i].AnimCommand);
         }
      fprintf(fp, "NumStateChanges: %d  (", Level->NumStateChanges);
      if (Level->NumStateChanges > 0) {
            for (i = 0, j = 0, k = 0x7fffffff; i < (int)Level->NumStateChanges; ++i) {
               if (Level->StateChanges[i].StateID > j)
                  j = Level->StateChanges[i].StateID;
               if (Level->StateChanges[i].StateID < k)
                  k = Level->StateChanges[i].StateID;
               }
            fprintf(fp, "MinMax %d..%d", k, j);
         }
      fprintf(fp, ") -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n");
      for (i = 0; i < (int)Level->NumStateChanges; ++i) {
         fprintf(fp, "   StateChanges[%3d]: StateID %3d NumAnimDispatches %2d AnimDispatch %d\n", i,
                 Level->StateChanges[i].StateID,
                 Level->StateChanges[i].NumAnimDispatches,
                 Level->StateChanges[i].AnimDispatch);
         }
      fprintf(fp, "NumAnimDispatches: %d (", Level->NumAnimDispatches);
      if (Level->NumAnimDispatches > 0) {
         for (i = 0, j = 0, k = 0x7fffffff; i < (int)Level->NumAnimDispatches; ++i) {
            if (Level->AnimDispatches[i].Low > j)
               j = Level->AnimDispatches[i].Low;
            if (Level->AnimDispatches[i].Low < k)
               k = Level->AnimDispatches[i].Low;
           }
         fprintf(fp, "MinMax %d..%d", k, j);
         }
      fprintf(fp, ") -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n");
      for (i = 0; i < (int)Level->NumAnimDispatches; ++i) {
         fprintf(fp, "   AnimDispatches[%3d]: Low %5d  High %5d  NextAnimation %5d  NextFrame %5d\n", i,
                 Level->AnimDispatches[i].Low,
                 Level->AnimDispatches[i].High,
                 Level->AnimDispatches[i].NextAnimation,
                 Level->AnimDispatches[i].NextFrame);
         }
      for (i = 0, j = 0, k = 0x7fffffff; i < (int)Level->NumAnimCommands; ++i) {
         if (Level->AnimCommands[i].Value > j)
            j = Level->AnimCommands[i].Value;
         if (Level->AnimCommands[i].Value < k)
            k = Level->AnimCommands[i].Value;
         }
      fprintf(fp, "NumAnimCommands: %d (min %d  max %d) -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n", Level->NumAnimCommands, k, j);
      for (i = 0; i < (int)Level->NumAnimCommands; ++i) {
         fprintf(fp, "   AnimCommands[%4d] Value %6d (0x%04x)\n", i, Level->AnimCommands[i].Value, (bitu16)Level->AnimCommands[i].Value);
         }
      fprintf(fp, "NumMeshTrees: %d -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n", Level->NumMeshTrees);
      for (i = 0; i < (int)(Level->NumMeshTrees / sizeof(tr2_meshtree)); ++i) {
         fprintf(fp, "   MeshTrees[%4d] Flags 0x%04x  x %d  y %d  z %d\n", i,
                 Level->MeshTrees[i].Flags,
                 Level->MeshTrees[i].x, Level->MeshTrees[i].y, Level->MeshTrees[i].z);
         }
      for (i = 0, j = 0, k = 0x7fffffff; i < (int)Level->NumFrames; ++i) {
         if (Level->Frames[i] > j)
            j = Level->Frames[i];
         if (Level->Frames[i] < k)
            k = Level->Frames[i];
         }
      fprintf(fp, "NumFrames: %d (min %d  max %d) -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n", Level->NumFrames, k, j);
      for (i = 0; i < (int)Level->NumFrames; ++i) {
         fprintf(fp, "   Frames[%6d]: %5d (0x%04x) (%6d)\n", i, Level->Frames[i], Level->Frames[i], (bit16)Level->Frames[i]);
         }
      fprintf(fp, "NumMoveables: %d -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n", Level->NumMoveables);
      for (i = 0; i < (int)Level->NumMoveables; ++i) {
         fprintf(fp, "   Moveables[%3d]: ObjectID %d (0x%08x)\n", i,
                 Level->Moveables[i].ObjectID,
                 Level->Moveables[i].ObjectID);
         fprintf(fp, "            NumMeshes %d  StartingMesh %d  MeshTree %d\n",
                 Level->Moveables[i].NumMeshes,
                 Level->Moveables[i].StartingMesh,
                 Level->Moveables[i].MeshTree);
         fprintf(fp, "            FrameOffset %d  Animation %d\n",
                 Level->Moveables[i].FrameOffset,
                 Level->Moveables[i].Animation);
         }
      fprintf(fp, "NumStaticMeshes: %d -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n", Level->NumStaticMeshes);
#ifdef NEVER_COMPILE_THIS
      for (i = 0; i < (int)Level->NumStaticMeshes; ++i) {
         fprintf(fp, "   StaticMeshes[%3d]: ObjectID %d  Unknown1 %d  StartingMesh %d\n", i,
                 Level->StaticMeshes[i].ObjectID, Level->StaticMeshes[i].Unknown1, Level->StaticMeshes[i].StartingMesh);
         for (j = 0; j < 13; ++j) {
            fprintf(fp, "      Unknown2[%2d]: %5d (0x%04x) (%6d)\n", j, Level->StaticMeshes[i].Unknown2[j], Level->StaticMeshes[i].Unknown2[j], (bit16)Level->StaticMeshes[i].Unknown2[j]);
            }
         }
#endif
      fprintf(fp, "NumObjectTextures: %d -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n", Level->NumObjectTextures);
      for (i = 0; i < (int)Level->NumObjectTextures; ++i) {
         fprintf(fp, "   ObjectTextures[%4d]: TransparencyFlags %d  Tile %d\n", i, Level->ObjectTextures[i].TransparencyFlags, Level->ObjectTextures[i].Tile);
         for (j = 0; j < 4; ++j) {
            fprintf(fp, "                 Vertex %2d: XCoord %3d  XPixel %3d  Ycoord %3d  Ypixel %3d\n", j,
                    Level->ObjectTextures[i].Vertices[j].Xcoordinate,
                    Level->ObjectTextures[i].Vertices[j].Xpixel,
                    Level->ObjectTextures[i].Vertices[j].Ycoordinate,
                    Level->ObjectTextures[i].Vertices[j].Ypixel);
            }
         }
      fprintf(fp, "NumSpriteTextures: %d -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n", Level->NumSpriteTextures);
      for (i = 0; i < (int)Level->NumSpriteTextures; ++i) {
         fprintf(fp, "   SpriteTextures[%3d]: Tile %d  X %d  Y %d\n", i,
                 Level->SpriteTextures[i].Tile,
                 Level->SpriteTextures[i].x,
                 Level->SpriteTextures[i].y);
         fprintf(fp, "                 Width %3d  Height %3d\n",
                 Level->SpriteTextures[i].Width,
                 Level->SpriteTextures[i].Height);
         fprintf(fp, "                 LeftSide %3d  TopSide %3d  RightSide %3d  BottomSide %3d\n",
                 Level->SpriteTextures[i].LeftSide,
                 Level->SpriteTextures[i].TopSide,
                 Level->SpriteTextures[i].RightSide,
                 Level->SpriteTextures[i].BottomSide);
         }
      fprintf(fp, "NumSpriteSequences: %d -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n", Level->NumSpriteSequences);
      for (i = 0; i < (int)Level->NumSpriteSequences; ++i) {
         fprintf(fp, "   SpriteSequences[%2d]: ObjectID %d (0x%08x)  NegativeLength %4d  Offset %3d (0x%04x)\n",
                 i, Level->SpriteSequences[i].ObjectID, Level->SpriteSequences[i].ObjectID, Level->SpriteSequences[i].NegativeLength, Level->SpriteSequences[i].Offset, (bitu16)Level->SpriteSequences[i].Offset);
         }
      fprintf(fp, "NumCameras: %d -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n", Level->NumCameras);
      for (i = 0; i < Level->NumCameras; ++i) {
         fprintf(fp, "   Cameras[%2d]: ", i);
         fprintf(fp, "X %d  Y %d  Z %d  Room %d  Index2 %d\n",
                 Level->Cameras[i].x, Level->Cameras[i].y, Level->Cameras[i].z, Level->Cameras[i].Room, Level->Cameras[i].Unknown1);
         }
      fprintf(fp, "NumSoundSources: %d -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n", Level->NumSoundSources);
      for (i = 0; i < Level->NumSoundSources; ++i) {
         fprintf(fp, "   SoundSources[%d]: x %ld  y %ld  z %ld  SoundMapIndex %d  Flags 0x%04x\n", i,
                 Level->SoundSources[i].x, Level->SoundSources[i].y, Level->SoundSources[i].z,
                 Level->SoundSources[i].SoundID, Level->SoundSources[i].Flags);
         }
      fprintf(fp, "NumBoxes: %d ", Level->NumBoxes);
      {
         int MinZ0 = 999,
             MinZ1 = 999,
             MaxZ0 = -999,
             MaxZ1 = -999,
             MinX0 = 999,
             MinX1 = 999,
             MaxX0 = -999,
             MaxX1 = -999;
         for (i = 0; i < Level->NumBoxes; ++i) {
            if (Level->Boxes[i].Zmin < MinZ0)
               MinZ0 = Level->Boxes[i].Zmin;
            if (Level->Boxes[i].Zmin > MaxZ0)
               MaxZ0 = Level->Boxes[i].Zmin;
            if (Level->Boxes[i].Zmax < MinZ1)
               MinZ1 = Level->Boxes[i].Zmax;
            if (Level->Boxes[i].Zmax > MaxZ1)
               MaxZ1 = Level->Boxes[i].Zmax;
            if (Level->Boxes[i].Xmin < MinX0)
               MinX0 = Level->Boxes[i].Xmin;
            if (Level->Boxes[i].Xmin > MaxX0)
               MaxX0 = Level->Boxes[i].Xmin;
            if (Level->Boxes[i].Xmax < MinX1)
               MinX1 = Level->Boxes[i].Xmax;
            if (Level->Boxes[i].Xmax > MaxX1)
               MaxX1 = Level->Boxes[i].Xmax;
            }
         fprintf(fp, "Zmin %d/%d  Zmax %d/%d  Xmin %d/%d  Xmax %d/%d ", MinZ0, MaxZ0, MinZ1, MaxZ1, MinX0, MaxX0, MinX1, MaxX1);
      }
      for (i = 0, j = 0, k = 0x7fffffff; i < Level->NumBoxes; ++i) {
         if (Level->Boxes[i].TrueFloor > j)
            j = Level->Boxes[i].TrueFloor;
         if (Level->Boxes[i].TrueFloor < k)
            k = Level->Boxes[i].TrueFloor;
         }
      fprintf(fp, "TrueFloor %d/%d ", k, j);
      {
      int BoxFloors[2048], BoxIx = 0, jjj;
      for (i = 0, l = 0; i < Level->NumBoxes; ++i) {
         int found = FALSE;
         for (jjj = 0; jjj < BoxIx && !found; ++jjj) {
            if (Level->Boxes[i].TrueFloor == BoxFloors[jjj])
               found = TRUE;
            }
         if (!found)
            BoxFloors[BoxIx++] = Level->Boxes[i].TrueFloor;
         }
      fprintf(fp, "(%d unique TrueFloors) ", BoxIx);
      }
      for (i = l, j = 0, k = 0x7fffffff; i < Level->NumBoxes; ++i) {
         if (Level->Boxes[i].OverlapIndex > j)
            j = Level->Boxes[i].OverlapIndex;
         if (Level->Boxes[i].OverlapIndex < k)
            k = Level->Boxes[i].OverlapIndex;
         }
      fprintf(fp, "OverlapIndex %d/%d ", k, j);
      fprintf(fp, "-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n");
      for (i = 0; i < Level->NumBoxes; ++i) {
         fprintf(fp, "   Boxes[%4d]: (Z %3d %3d, X %3d %3d)  TrueFloor %6d  OverlapIndex %5d", i,
                 Level->Boxes[i].Zmin, Level->Boxes[i].Zmax, Level->Boxes[i].Xmin, Level->Boxes[i].Xmax,
                 Level->Boxes[i].TrueFloor, Level->Boxes[i].OverlapIndex & 0x7fff);
         if (Level->Boxes[i].OverlapIndex & 0x8000)
            fprintf(fp, " | 0x8000");
         fprintf(fp, "\n");
         }
      for (i = 0, j = 0, k = 0x7fffffff; i < Level->NumOverlaps; ++i) {
         if ((Level->Overlaps[i] & 0x7fff) > j)
            j = Level->Overlaps[i] & 0x7fff;
         if ((Level->Overlaps[i] & 0x7fff) < k)
            k = Level->Overlaps[i] & 0x7fff;
         }
      if (Level->NumOverlaps == 0) {
         j = 0;
         k = 0;
         }
      fprintf(fp, "NumOverlaps: %d (min %d  max %d)", Level->NumOverlaps, k, j);
      j = 99999;
      k = 0;
      for (i = 0; i < Level->NumOverlaps; ) {
         l = 1;
         while (!(Level->Overlaps[i++] & 0x8000))
            ++l;
         if (l < j)
            j = l;
         if (l > k)
            k = l;
         }
      fprintf(fp, " (runs: shortest %d  longest %d) -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n", j, k);
      for (i = 0; i < Level->NumOverlaps; ++i) {
         fprintf(fp, "   Overlaps[%4d]: %6d (0x%04x) (%5d)\n", i, Level->Overlaps[i], (bitu16)Level->Overlaps[i], Level->Overlaps[i] & 0x7fff);
         }

      fprintf(fp, "Zones: MinMax ");
      for (l = 0; l < 10; ++l) {
         for (i = 0, j = 0, k = 0x7fffffff; i < Level->NumBoxes; ++i) {
            if (Level->Zones[(l * Level->NumBoxes) + i] > j)
               j = Level->Zones[(l * Level->NumBoxes) + i];
            if (Level->Zones[(l * Level->NumBoxes) + i] < k)
               k = Level->Zones[(l * Level->NumBoxes) + i];
            }
         fprintf(fp, "%3d/%3d ", k, j);
         }
      fprintf(fp, " Num %d -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n", Level->NumBoxes);

      for (i = 0; i < Level->NumBoxes; ++i) {
        fprintf(fp, "   Zones[%4d]: ", i);
        for (j = 0; j < 10; ++j) {
           fprintf(fp, "%6d ", Level->Zones[(j * Level->NumBoxes) + i]);
           }
        fprintf(fp, "\n");
        }
      fprintf(fp, "NumAnimatedTextures: %d -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n", Level->NumAnimatedTextures);
      for (i = 0; i < Level->NumAnimatedTextures; ++i) {
         fprintf(fp, "   AnimatedTextures[%2d]: %6d (0x%04x)\n", i, Level->AnimatedTextures[i], (bitu16)Level->AnimatedTextures[i]);
         }
      fprintf(fp, "NumItems: %d -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n", Level->NumItems);
      for (i = 0; i < Level->NumItems; ++i) {
         fprintf(fp, "   Items[%3d]: ObjectID %d (0x%04x)  Room %d\n", i, Level->Items[i].ObjectID, Level->Items[i].ObjectID, Level->Items[i].Room);
         fprintf(fp, "        position (%5d, %5d, %5d)\n", Level->Items[i].x, Level->Items[i].y, Level->Items[i].z);
         fprintf(fp, "        Angle %d  Intensity1 %d  Intensity2 %d  Flags %d\n",
                 Level->Items[i].Angle,
                 Level->Items[i].Intensity1,
                 Level->Items[i].Intensity2,
                 Level->Items[i].Flags);
         }
      fprintf(fp, "LightMap (32 * 256) -=-=-=-=-=-=-=-=-=-=-=-\n");
      for (i = 0; i < 32; ++i) {
         fprintf(fp, "LightMap[%d]:\n", i);
         HexDump(&Level->LightMap[i * 256], 256, 0, "   ", fp);
         }
      fprintf(fp, "NumCinematicFrames: %d (%d bytes) -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n", Level->NumCinematicFrames, Level->NumCinematicFrames * 16);
      for (i = 0; i < Level->NumCinematicFrames; ++i) {
         fprintf(fp, "   CinematicFrames[%3d]: ", i);
         HexDump((bitu8 *)&Level->CinematicFrames[i * 8], 16, i * 16, "", fp);
         fprintf(fp, "           [%3d]: %6d %6d %6d %6d %6d %6d %6d %6d\n", i,
                 Level->CinematicFrames[(i * 8) + 0],
                 Level->CinematicFrames[(i * 8) + 1],
                 Level->CinematicFrames[(i * 8) + 2],
                 Level->CinematicFrames[(i * 8) + 3],
                 Level->CinematicFrames[(i * 8) + 4],
                 Level->CinematicFrames[(i * 8) + 5],
                 Level->CinematicFrames[(i * 8) + 6],
                 Level->CinematicFrames[(i * 8) + 7]);
         }
      fprintf(fp, "NumDemoData: %d -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n", Level->NumDemoData);
      HexDump(Level->DemoData, Level->NumDemoData, 0, "   ", fp);
      fprintf(fp, "SoundMap -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n");
      for (i = 0; i < 370; ++i)
         fprintf(fp, "   SoundMap[%3d] = %d\n", i, Level->SoundMap[i]);

      fprintf(fp, "NumSoundDetails: %d -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n", Level->NumSoundDetails);
      for (i = 0; i < Level->NumSoundDetails; ++i) {
         fprintf(fp, "   SoundDetails[%3d]: Sample %3d  Volume %d  Unknown1 %6d  Unknown2 0x%04x (flags1 %d / numsamp %d / flags2 %d)\n", i,
                 Level->SoundDetails[i].Sample, Level->SoundDetails[i].Volume, Level->SoundDetails[i].SoundRange,
                 Level->SoundDetails[i].Flags, (Level->SoundDetails[i].Flags & 0xf000) >> 24,
                 (Level->SoundDetails[i].Flags & 0x0ffc) >> 2, Level->SoundDetails[i].Flags & 0x0003);
         }
      fprintf(fp, "NumSampleIndices: %d -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n", Level->NumSampleIndices);
      for (i = 0; i < Level->NumSampleIndices; ++i) {
         fprintf(fp, "   SampleIndices[%3d]: %6d (0x%08x)\n", i, Level->SampleIndices[i], Level->SampleIndices[i]);
         }

      fclose(fp);
      ReturnValue = 0;
      }
   else {
      perror(FileName);
      ReturnValue = -1;
      }
   return(ReturnValue);
}


