
/*
 * tr2rd.c
 *
 * Tomb Raider II read routines for TRueView
 *
 *** CURRENTLY LACKING:
 ***     Better error control (specifically, any malloc() or fread() error
 ***                           just unceremoniously exits the program)
 ***     Correct handling for unknowns (obviously, I can't correctly
 ***                           handle what I don't understand)
 */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <memory.h>
#include <string.h>
#include <stdlib.h>

#include "config.h"

#ifdef INCLUDE_TR4_SUPPORT
// ZLIB #defines and #include for TR4 compression support
#ifndef LINUX
#define _WINDOWS 1
#define ZLIB_DLL 1
#endif // LINUX
#include "zlib.h"
#endif // INCLUDE_TR4_SUPPORT

#include "tr2port.h"
#include "tr2.h"

#ifdef RWTEST
#define DUMPRD(f, s)	ioDumpRd(f, s)
#define SAVERD(f, b, s)	ioSaveRd(f, b, s)
#else
#define DUMPRD(f, s)
#define SAVERD(f, b, s)
#endif

/*
 * ioCopyRd -- copy file sections after reading
 */

#ifdef RWTEST
static long
ioCopyRd(tr2_file *fp, long offset, char *name)
{
   static int count = 0;
   char fname[64];
   FILE *ffp;
   tr2_file td, *dp = &td;

   /* open dump file */
   sprintf(fname, "dumprd-%03d-%s", count++, name);
   if ((ffp = fopen(fname, "w")) == NULL) {
      perror("ioCopyRd: fopen failed");
      exit(2);
   }
   TR2fmake(dp, TombRaider_2, ffp, NULL, 0);
 
   /* copy stuff */
   offset = TR2fcopy(fp, offset, dp);
 
   /* close dump file */
   if (fclose(ffp) != 0) {
      perror("ioCopyRd: fclose failed");
      exit(2);
   } 

   return(offset);
}

/*
 * ioDumpRd -- dump file sections after reading
 */

static void
ioDumpRd(tr2_file *fp, char *name)
{
   static int filemode = 1;
   static long fileoff = 0, buffoff = 0;

   /* initialize */
   if (name == NULL) {
      filemode = (fp->Fp != NULL);
      if (filemode) fileoff = 0;
      else buffoff = fp->DataOffset;
      return;
   }

   /* change mode */
   if (filemode && (fp->Fp == NULL)) {
      buffoff = 0;
   }
   if (!filemode && (fp->Fp != NULL)) {
      fileoff += buffoff; buffoff = 0;
   }
   filemode = (fp->Fp != NULL);

   /* dump stuff */
   if (filemode) fileoff = ioCopyRd(fp, fileoff, name);
   else          buffoff = ioCopyRd(fp, buffoff, name);

   return;
}

/*
 * ioSaveRd -- save file sections after reading
 */

static void
ioSaveRd(tr2_file *fp, long offset, char *name)
{
   static tr2_file tb = {0}, *bp = &tb;

   if (fp != NULL) {
      TR2fcopy(fp, offset, bp);
   }

   if (name != NULL) {
      ioCopyRd(bp, 0, name);
      FreeIfUsed(bp->Data); bp->Data = NULL;
      bp->DataOffset = bp->DataSize = 0;
   }

   return;
}
#endif  /* RWTEST */

/*
 * TR2_Version_type TR2getVersion(bitu32 version)
 *
 * Returns TombRaider_1..TombRaider_3 or TombRaider_VersionUnknown based on <version>
 */
TR2_Version_type TR2getVersion(bitu32 version)
{
   int i;
   static struct {
      bitu32 LongVersion;
      TR2_Version_type Version;
      } TR2_Versions[] = {
         { 0x00000020, TombRaider_1 },
         { 0x0000002d, TombRaider_2 },
         { 0xff080038, TombRaider_3 },
         { 0xff180038, TombRaider_3 },
         { 0xfffffff0, TombRaider_4 }, // bogus
         { 0x00345254, TombRaider_4 }, // "TR4\0"
         { 0x00000000, TombRaider_UnknownVersion }
      };

   for (i = 0; TR2_Versions[i].LongVersion != 0; ++i) {
      if (TR2_Versions[i].LongVersion == version)
         break;
      }
   return(TR2_Versions[i].Version);
}

/*
 * MapTR1ItemToTR2Item()
 *
 * Does what its name implies - maps TR1 ItemIDs to TR2 ItemIDs
 */

#ifdef NEVER_COMPILE_THIS
static int MapTR1ItemToTR2Item(int ItemID)
{
   int ReturnValue;
   static int TR1toTR2map[300] = {
      0,    //   0:0 Lara
      1,    //   1:1 Lara pistol animation
      3,    //   2:3 Lara shotgun animation
      4,    //   3:4 Lara magnum animation
      5,    //   4:5 Lara Uzi animation
      0,    //   5:? Lara / Lara's home appearance / Lara wounded / Lara turned to gold
      0,    //   6:? Lara's evil twin mutant
      36,   //   7:36 (15?) Wolf
      37,   //   8:37 Bear
      38,   //   9:38? Bat
      0,    //  10: Crocodile (on land)
      0,    //  11: Crocodile (in water)
      39,   //  12:39? Lion (male)
      39,   //  13:39? Lion (female)
      39,   //  14:39? Panther
      45,   //  15:45? Gorilla
      21,   //  16:21? Giant Rat (on land)
      21,   //  17:21? Giant Rat (in water)
      214,  //  18:214 Tyrannosaur
      0,    //  19:? Raptor
      0,    //  20: Winged mummy (unused) / Winged mutant
      0,    //  21: Lara's hips
      0,    //  22: Lara's hips
      0,    //  23: Centaur mutant
      0,    //  24: Mummy
      0,    //  25: ?
      0,    //  26: ?
      0,    //  27: Larson
      0,    //  28:? (can't die until later) Pierre
      0,    //  29: Skateboard
      0,    //  30: Skateboard kid
      0,    //  31: Cowboy
      0,    //  32: "Mr. T"
      0,    //  33: Winged Natla (actually, Natla with a winged mutant)
      0,    //  34: Giant mutant
      55,   //  35:55 Collapsible floor
      0,    //  36: Swinging blade
      59,   //  37:59 Spikes
      60,   //  38:60 Boulder
      61,   //  39:61 Dart
      62,   //  40:62 Wall-mounted dartgun
      63,   //  41:63? Door (opens upward)
      64,   //  42:64 Slamming doors
      0,    //  43: Sword of Damocles
      0,    //  44: Thor's hammer's handle
      0,    //  45: Thor's hammer's block
      0,    //  46: Hanging ball? / Some kind of box?
      0,    //  47: Metal rod? / Powered mining cart
      67,   //  48:67 Movable cubical block (pushable)
      68,   //  49:68 Movable cubical block (pushable)
      69,   //  50:69 Movable cubical block (pushable)
      70,   //  51:70 Movable cubical block (pushable)
      0,    //  52: Movable tall block
      0,    //  53: Pieces of something?
      0,    //  54: Sword of Damocles
      103,  //  55:103 Above-water switch
      105,  //  56:105 Underwater switch
      106,  //  57:106 Door
      107,  //  58:107 Door
      108,  //  59:108 Door
      109,  //  60:109 Door
      110,  //  61:110 Door
      111,  //  62:111 Door
      112,  //  63:112 Door
      113,  //  64:113 Door
      0,    //  65: Trapdoor (opens downward)
      0,    //  66: Trapdoor (opens downward)
      0,    //  67: ?
      117,  //  68:117 Bridge (flat)
      118,  //  69:118 Bridge (slope = 1)
      119,  //  70:119 Bridge (slope = 2)
      120,  //  71:120 Passport (opening up)
      121,  //  72:121? Compass
      0,    //  73: ?
      0,    //  74: Cogs (animated)
      0,    //  75: Cogs (animated)
      0,    //  76: Cogs (animated)
      0,    //  77: Lara in CS / Scion holder in CS
      0,    //  78: Larson in CS / Natla in CS / Scion holder in CS
      0,    //  79: Larson's gun in CS / Scion in CS / Natla in CS
      0,    //  80: Scion in CS
      133,  //  81:133 Passport (closed)
      134,  //  82:134 N-thingy (Playstation memory card?)
      0,    //  83: Save crystal
      135,  //  84:135 * Pistols
      136,  //  85:136 * Shotgun
      137,  //  86:137 * Magnums
      138,  //  87:138 * Uzis
      0,    //  88:? * Pistol ammo(?)
      143,  //  89:143 * Shotgun ammo
      144,  //  90:144 * Magnum ammo
      145,  //  91:145 * Uzi ammo
      0,    //  92: ?
      149,  //  93:149 * Small medipack
      150,  //  94:150 * Large medipack
      153,  //  95:153 Sunglasses
      154,  //  96:154 Cassette player and headphones
      155,  //  97:155 Direction keys
      0,    //  98: ?
      157,  //  99:157 Pistol
      158,  // 100:158 Shotgun
      159,  // 101:159 Magnum
      160,  // 102:160 Uzi
      164,  // 103:164 Pistol ammo(?)
      165,  // 104:165 Shotgun ammo
      166,  // 105:166 Magnum ammo
      167,  // 106:167 Uzi ammo
      0,    // 107: ?
      171,  // 108:171 Small medipack
      172,  // 109:172 Large medipack
      174,  // 110:174 * Key 1
      175,  // 111:175 * Key 2
      176,  // 112:176 * Key 3
      177,  // 113:177 * Key 4
      178,  // 114:178 Key 1
      179,  // 115:179 Key 2
      180,  // 116:180 Key 3
      181,  // 117:181 Key 4
      182,  // 118:182 Lock 1 empty
      183,  // 119:183 Lock 2 empty
      184,  // 120:184 Lock 3 empty
      185,  // 121:185 Lock 4 empty
      186,  // 122:186 Lock 1 full
      187,  // 123:187 Lock 2 full
      188,  // 124:188 Lock 3 full
      189,  // 125:189 Lock 4 full
      205,  // 126:205 * Key 5
      207,  // 127:207 Key 5
      0,    // 128: Lara's hips
      206,  // 129:206 * Key 6
      193,  // 130:193 * Key 7
      194,  // 131:194 * Key 8
      195,  // 132:195 * Key 9
      208,  // 133:208 Key 6
      197,  // 134:197 Key 7
      198,  // 135:198 Key 8
      199,  // 136:199 Key 9
      0,    // 137:? Lock 6
      201,  // 138:201 Lock 7
      202,  // 139:202 Lock 8
      203,  // 140:203 Lock 9
      0,    // 141: ?
      0,    // 142: ?
      0,    // 143: * Scion Piece
      0,    // 144: ?
      0,    // 145: ?
      0,    // 146: Complete Scion
      0,    // 147: Scion Holder
      0,    // 148: ?
      0,    // 149: ?
      0,    // 150: Scion Piece
      229,  // 151:229 * Flare(?) / Explosion
      0,    // 152: ?
      230,  // 153:230 * Splash
      0,    // 154: ?
      231,  // 155:231 * Bubbles
      231,  // 156:231? * Bubbles
      0,    // 157: ?
      233,  // 158:233 * Blood splatter
      0,    // 159: ?
      234,  // 160:234? * Flying disk
      0,    // 161: Centaur statue
      0,    // 162: Shack suspended from wire rope
      0,    // 163: Mutant egg and holder (normal size)
      238,  // 164:238 * Bullet hit
      239,  // 165:239 * Sparkle
      240,  // 166:240 Gunflare
      0,    // 167: ?
      0,    // 168: ?
      0,    // 169: Lara's hips
      0,    // 170: Lara's hips
      0,    // 171: ?
      0,    // 172: Mutant bullet
      0,    // 173: Mutant grenade
      0,    // 174: ?
      0,    // 175: ?
      0,    // 176: * Splatter
      0,    // 177: Lara's hips
      252,  // 178:252 * Fire
      0,    // 179: Lara's hips
      0,    // 180: Flowing Atlantean lava
      0,    // 181: Mutant egg and holder (big)
      14,   // 182:14? Motorboat
      0,    // 183: Lara's hips
      0,    // 184: ?
      0,    // 185: ?
      0,    // 186: ?
      0,    // 187: ?
      0,    // 188: ?
      0,    // 189: Shrinking wedge?
      255,  // 190:255 * Standard symbols
      0,    // 191: * Plant 1
      0,    // 192: * Plant 2
      0,    // 193: * Plant 3
      0,    // 194: * Plant 4
      0,    // 195: * Plant 5
      0,    // 196: ?
      0,    // 197: ?
      0,    // 198: ?
      0,    // 199: ?
      0,    // 200: * Bag 1
      0,    // 201: ?
      0,    // 202: ?
      0,    // 203: ?
      0,    // 204: * Bag 2
      0,    // 205: ?
      0,    // 206: ?
      241,  // 207:241 Gunflare
      0,    // 208: ?
      0,    // 209: ?
      0,    // 210: ?
      0,    // 211: ?
      0,    // 212: * Rock 1
      0,    // 213: * Rock 2
      0,    // 214: * Rock 3
      0,    // 215: * Bag 3
      0,    // 216: * Pottery 1
      0,    // 217: * Pottery 2
      0,    // 218: ?
      0,    // 219: ?
      0,    // 220: ?
      0,    // 221: ?
      0,    // 222: ?
      0,    // 223: ?
      0,    // 224: ?
      0,    // 225: ?
      0,    // 226: ?
      0,    // 227: ?
      0,    // 228: ?
      0,    // 229: ?
      0,    // 230: ?
      0,    // 231: * Painted pot
      0,    // 232: ?
      0,    // 233: * Inca mummy
      0,    // 234: ?
      0,    // 235: ?
      0,    // 236: * Pottery 3
      0,    // 237: * Pottery 4
      0,    // 238: * Pottery 5
      0,    // 239: * Pottery 6
      };

   ReturnValue = TR1toTR2map[ItemID];
   return((ReturnValue == 0 && ItemID != 0) ? ItemID : ReturnValue);
}
#endif /* NEVER_COMPILE_THIS */

/*
 * ioConvertTR1Frames --
 * re-format TR1 Frames[] to look like TR2 frames
 */

static void
ioConvertTR1Frames(
   TR2_Level *Level)
{
   int Num, i, j;

   for (j = 0; j < (int)Level->NumAnimations; ++j) {
      int fo = Level->Animations[j].FrameOffset / 2;
      Level->Animations[j].FrameSize = (Level->Frames[fo + 9] * 2) + 10;
      }

   /*
    * Turn    . . . . . . . . . n 1 2 3 4 5 6    (n = Num := 3)
    * into    . . . . . . . . . 2 1 4 3 6 5 6
    *
    * Each pair of frames is reversed and shifted to the left,
    * the last frame is duplicated. Num is overwritten; the
    * original value was stored in Animation.FrameSize above.
    */

   for (i = 0; i < (int)Level->NumFrames; ) {
      i += 9;  // point to Num;
      Num = Level->Frames[i];
      for (; Num != 0; Num--, i+=2) {
         Level->Frames[i] = Level->Frames[i+2];
         }
      i++;
      }

   return;
}

/*
 * TR2fread_xxx -- read miscellaneous data
 */

static int
TR2fread_colour(
  tr2_colour *buf, size_t count, tr2_file *fp)
{
  int NumRead = 0, i;

  for (i = 0; i < count; i++) {
    NumRead += TR2fread1_bitu8(&buf[i].Red, fp);
    NumRead += TR2fread1_bitu8(&buf[i].Green, fp);
    NumRead += TR2fread1_bitu8(&buf[i].Blue, fp);
  }
  return(NumRead);
}

static int
TR2fread_vertex(
  tr2_vertex *buf, size_t count, tr2_file *fp)
{
  int NumRead = 0, i;

  for (i = 0; i < count; i++) {
    NumRead += TR2fread1_bit16(&buf[i].x, fp);
    NumRead += TR2fread1_bit16(&buf[i].y, fp);
    NumRead += TR2fread1_bit16(&buf[i].z, fp);
  }
  return(NumRead);
}

static int
TR2fread_face4(
  tr2_face4 *buf, size_t count, tr2_file *fp)
{
  int NumRead = 0, i;

  for (i = 0; i < count; i++) {
    NumRead += TR2fread_bitu16(buf[i].Vertices, dimof(buf[i].Vertices), fp);
    NumRead += TR2fread1_bitu16(&buf[i].Texture, fp);
#ifndef RWTEST
    if (fp->Version >= TombRaider_3) buf[i].Texture &= 0x7fff;
#endif
  }
  return(NumRead);
}

static int
TR2fread_face4c(
  tr2_face4 *buf, size_t count, tr2_file *fp)
{
  int NumRead = 0, i;

  for (i = 0; i < count; i++) {
    NumRead += TR2fread_bitu16(buf[i].Vertices, dimof(buf[i].Vertices), fp);
    NumRead += TR2fread1_bitu16(&buf[i].Texture, fp);
#ifndef RWTEST
    if (fp->Version >= TombRaider_3) buf[i].Texture &= 0x7fff;
#endif
    /* TR1: add index into 16-bit palette */
    if (fp->Version == TombRaider_1) {
      buf[i].Texture = (buf[i].Texture & 0xff);
      buf[i].Texture |= (buf[i].Texture << 8);
    }
  }
  return(NumRead);
}

static int
TR2fread_face3(
  tr2_face3 *buf, size_t count, tr2_file *fp)
{
  int NumRead = 0, i;

  for (i = 0; i < count; i++) {
    NumRead += TR2fread_bitu16(buf[i].Vertices, dimof(buf[i].Vertices), fp);
    NumRead += TR2fread1_bitu16(&buf[i].Texture, fp);
#ifndef RWTEST
    if (fp->Version >= TombRaider_3) buf[i].Texture &= 0x7fff;
#endif
  }
  return(NumRead);
}

static int
TR2fread_face3c(
  tr2_face3 *buf, size_t count, tr2_file *fp)
{
  int NumRead = 0, i;

  for (i = 0; i < count; i++) {
    NumRead += TR2fread_bitu16(buf[i].Vertices, dimof(buf[i].Vertices), fp);
    NumRead += TR2fread1_bitu16(&buf[i].Texture, fp);
#ifndef RWTEST
    if (fp->Version >= TombRaider_3) buf[i].Texture &= 0x7fff;
#endif
    /* TR1: add index into 16-bit palette */
    if (fp->Version == TombRaider_1) {
      buf[i].Texture = (buf[i].Texture & 0xff);
      buf[i].Texture |= (buf[i].Texture << 8);
    }
  }
  return(NumRead);
}

static int
TR2fread_textile8(
  tr2_textile8 *buf, size_t count, tr2_file *fp)
{
  int NumRead = 0, i;

  for (i = 0; i < count; i++) {
    NumRead += TR2fread_bitu8(buf[i].Tile, dimof(buf[i].Tile), fp);
  }
  return(NumRead);
}

static int
TR2fread_textile16(
  tr2_textile16 *buf, size_t count, tr2_file *fp)
{
  int NumRead = 0, i;

  for (i = 0; i < count; i++) {
    NumRead += TR2fread_bitu16(buf[i].Tile, dimof(buf[i].Tile), fp);
  }
  return(NumRead);
}

static int
TR2fread_room_info(
  tr2_room_info *buf, size_t count, tr2_file *fp)
{
  int NumRead = 0, i;

  for (i = 0; i < count; i++) {
    NumRead += TR2fread1_bit32(&buf[i].x, fp);
    NumRead += TR2fread1_bit32(&buf[i].z, fp);
    NumRead += TR2fread1_bit32(&buf[i].yBottom, fp);
    NumRead += TR2fread1_bit32(&buf[i].yTop, fp);
  }
  return(NumRead);
}

static int
TR2fread_room_portal(
  tr2_room_portal *buf, size_t count, tr2_file *fp)
{
  int NumRead = 0, i;

  for (i = 0; i < count; i++) {
    NumRead += TR2fread1_bitu16(&buf[i].AdjoiningRoom, fp);
    NumRead += TR2fread_vertex(&buf[i].Normal, 1, fp);
    NumRead += TR2fread_vertex(buf[i].Vertices, dimof(buf[i].Vertices), fp);
  }
  return(NumRead);
}

static int
TR2fread_room_sector(
  tr2_room_sector *buf, size_t count, tr2_file *fp)
{
  int NumRead = 0, i;

  for (i = 0; i < count; i++) {
    NumRead += TR2fread1_bitu16(&buf[i].FDindex, fp);
    NumRead += TR2fread1_bitu16(&buf[i].BoxIndex, fp);
    NumRead += TR2fread1_bitu8(&buf[i].RoomBelow, fp);
    NumRead += TR2fread1_bit8(&buf[i].Floor, fp);
    NumRead += TR2fread1_bitu8(&buf[i].RoomAbove, fp);
    NumRead += TR2fread1_bit8(&buf[i].Ceiling, fp);
  }
  return(NumRead);
}

static int
TR2fread_room_light(
  tr2_room_light *buf, size_t count, tr2_file *fp)
{
  int NumRead = 0, i;

  for (i = 0; i < count; i++) {
    NumRead += TR2fread1_bit32(&buf[i].x, fp);
    NumRead += TR2fread1_bit32(&buf[i].y, fp);
    NumRead += TR2fread1_bit32(&buf[i].z, fp);
    NumRead += TR2fread1_bitu16(&buf[i].Intensity1, fp);
    if (fp->Version == TombRaider_1) buf[i].Intensity2 = buf[i].Intensity1;
    else NumRead += TR2fread1_bitu16(&buf[i].Intensity2, fp);
    NumRead += TR2fread1_bitu32(&buf[i].Fade1, fp);
    if (fp->Version == TombRaider_1) buf[i].Fade2 = buf[i].Fade1;
    else NumRead += TR2fread1_bitu32(&buf[i].Fade2, fp);
    if (fp->Version != TombRaider_4) /*nop*/;
    else NumRead += TR2fread_bitu8(buf[i].junk, dimof(buf[i].junk), fp);
  }
  return(NumRead);
}

static int
TR2fread_vertex_room(
  tr2_vertex_room *buf, size_t count, tr2_file *fp)
{
  int NumRead = 0, i;

  for (i = 0; i < count; i++) {
    NumRead += TR2fread_vertex(&buf[i].Vertex, 1, fp);
    NumRead += TR2fread1_bit16(&buf[i].Lighting1, fp);
    if (fp->Version == TombRaider_1) buf[i].Attributes = 0;
    else NumRead += TR2fread1_bitu16(&buf[i].Attributes, fp);
    if (fp->Version == TombRaider_1) buf[i].Lighting2 = buf[i].Lighting1;
    else NumRead += TR2fread1_bit16(&buf[i].Lighting2, fp);
  }
  return(NumRead);
}

static int
TR2fread_room_sprite(
  tr2_room_sprite *buf, size_t count, tr2_file *fp)
{
  int NumRead = 0, i;

  for (i = 0; i < count; i++) {
    NumRead += TR2fread1_bit16(&buf[i].Vertex, fp);
    NumRead += TR2fread1_bit16(&buf[i].Texture, fp);
    if (fp->Version >= TombRaider_3) buf[i].Texture &= 0x7fff;
  }
  return(NumRead);
}

static int
TR2fread_room_staticmesh(
  tr2_room_staticmesh *buf, size_t count, tr2_file *fp)
{
  int NumRead = 0, i;

  for (i = 0; i < count; i++) {
    NumRead += TR2fread1_bit32(&buf[i].x, fp);
    NumRead += TR2fread1_bit32(&buf[i].y, fp);
    NumRead += TR2fread1_bit32(&buf[i].z, fp);
    NumRead += TR2fread1_bitu16(&buf[i].Rotation, fp);
    NumRead += TR2fread1_bitu16(&buf[i].Intensity1, fp);
    if (fp->Version == TombRaider_1) buf[i].Intensity2 = buf[i].Intensity1;
    else NumRead += TR2fread1_bitu16(&buf[i].Intensity2, fp);
    NumRead += TR2fread1_bitu16(&buf[i].ObjectID, fp);
  }
  return(NumRead);
}

static int
TR2fread_animation(
  tr2_animation *buf, size_t count, tr2_file *fp)
{
  int NumRead = 0, i;

  for (i = 0; i < count; i++) {
    NumRead += TR2fread1_bitu32(&buf[i].FrameOffset, fp);
    NumRead += TR2fread1_bitu8(&buf[i].FrameRate, fp);
    /* TR1: FrameSize is 0, will be overwritten */
    NumRead += TR2fread1_bitu8(&buf[i].FrameSize, fp);
    NumRead += TR2fread1_bit16(&buf[i].StateID, fp);
    NumRead += TR2fread1_bit16(&buf[i].Unknown1, fp);
    NumRead += TR2fread1_bit16(&buf[i].Unknown2, fp);
    NumRead += TR2fread1_bit16(&buf[i].Unknown3, fp);
    NumRead += TR2fread1_bit16(&buf[i].Unknown4, fp);
    if (fp->Version != TombRaider_4) /*nop*/;
    else NumRead += TR2fread_bitu8(buf[i].junk, dimof(buf[i].junk), fp);
    NumRead += TR2fread1_bitu16(&buf[i].FrameStart, fp);
    NumRead += TR2fread1_bitu16(&buf[i].FrameEnd, fp);
    NumRead += TR2fread1_bitu16(&buf[i].NextAnimation, fp);
    NumRead += TR2fread1_bitu16(&buf[i].NextFrame, fp);
    NumRead += TR2fread1_bitu16(&buf[i].NumStateChanges, fp);
    NumRead += TR2fread1_bitu16(&buf[i].StateChangeOffset, fp);
    NumRead += TR2fread1_bitu16(&buf[i].NumAnimCommands, fp);
    NumRead += TR2fread1_bitu16(&buf[i].AnimCommand, fp);
  }
  return(NumRead);
}

static int
TR2fread_state_change(
  tr2_state_change *buf, size_t count, tr2_file *fp)
{
  int NumRead = 0, i;

  for (i = 0; i < count; i++) {
    NumRead += TR2fread1_bitu16(&buf[i].StateID, fp);
    NumRead += TR2fread1_bitu16(&buf[i].NumAnimDispatches, fp);
    NumRead += TR2fread1_bitu16(&buf[i].AnimDispatch, fp);
  }
  return(NumRead);
}

static int
TR2fread_anim_dispatch(
  tr2_anim_dispatch *buf, size_t count, tr2_file *fp)
{
  int NumRead = 0, i;

  for (i = 0; i < count; i++) {
    NumRead += TR2fread1_bit16(&buf[i].Low, fp);
    NumRead += TR2fread1_bit16(&buf[i].High, fp);
    NumRead += TR2fread1_bit16(&buf[i].NextAnimation, fp);
    NumRead += TR2fread1_bit16(&buf[i].NextFrame, fp);
  }
  return(NumRead);
}

static int
TR2fread_anim_command(
  tr2_anim_command *buf, size_t count, tr2_file *fp)
{
  int NumRead = 0, i;

  for (i = 0; i < count; i++) {
    NumRead += TR2fread1_bit16(&buf[i].Value, fp);
  }
  return(NumRead);
}

static int
TR2fread_meshtree(
  tr2_meshtree *buf, size_t count, tr2_file *fp)
{
  int NumRead = 0, i;

  for (i = 0; i < count; i++) {
    NumRead += TR2fread1_bit32(&buf[i].Flags, fp);
    NumRead += TR2fread1_bit32(&buf[i].x, fp);
    NumRead += TR2fread1_bit32(&buf[i].y, fp);
    NumRead += TR2fread1_bit32(&buf[i].z, fp);
  }
  return(NumRead);
}

static int
TR2fread_moveable(
  tr2_moveable *buf, size_t count, tr2_file *fp)
{
  int NumRead = 0, i;

  for (i = 0; i < count; i++) {
    NumRead += TR2fread1_bitu32(&buf[i].ObjectID, fp);
    NumRead += TR2fread1_bitu16(&buf[i].NumMeshes, fp);
    NumRead += TR2fread1_bitu16(&buf[i].StartingMesh, fp);
    NumRead += TR2fread1_bitu32(&buf[i].MeshTree, fp);
    NumRead += TR2fread1_bitu32(&buf[i].FrameOffset, fp);
    NumRead += TR2fread1_bitu16(&buf[i].Animation, fp);
  }
  return(NumRead);
}

static int
TR2fread_staticmesh(
  tr2_staticmesh *buf, size_t count, tr2_file *fp)
{
  int NumRead = 0, i;

  for (i = 0; i < count; i++) {
    NumRead += TR2fread1_bitu32(&buf[i].ObjectID, fp);
    NumRead += TR2fread1_bitu16(&buf[i].StartingMesh, fp);
    NumRead += TR2fread_vertex(&buf[i].BoundingBox[0][0], dim2of(buf[i].BoundingBox), fp);
    NumRead += TR2fread1_bitu16(&buf[i].Flags, fp);
  }
  return(NumRead);
}

static int
TR2fread_object_texture_vert(
  tr2_object_texture_vert *buf, size_t count, tr2_file *fp)
{
  int NumRead = 0, i;

  for (i = 0; i < count; i++) {
    NumRead += TR2fread1_bitu8(&buf[i].Xcoordinate, fp);
    NumRead += TR2fread1_bitu8(&buf[i].Xpixel, fp);
    NumRead += TR2fread1_bitu8(&buf[i].Ycoordinate, fp);
    NumRead += TR2fread1_bitu8(&buf[i].Ypixel, fp);
  }
  return(NumRead);
}

static int
TR2fread_object_texture(
  tr2_object_texture *buf, size_t count, tr2_file *fp)
{
  int NumRead = 0, i;
  bitu32 Tile;

  for (i = 0; i < count; i++) {
    NumRead += TR2fread1_bitu16(&buf[i].TransparencyFlags, fp);
    if (fp->Version == TombRaider_4) {
      NumRead += TR2fread1_bitu32(&Tile, fp); buf[i].Tile = Tile & 0x7fff; }
    else {
      NumRead += TR2fread1_bitu16(&buf[i].Tile, fp);
    }
    NumRead += TR2fread_object_texture_vert(buf[i].Vertices, dimof(buf[i].Vertices), fp);
    if (fp->Version != TombRaider_4) /*nop*/;
    else NumRead += TR2fread_bitu8(buf[i].filler, dimof(buf[i].filler), fp);
  }
  return(NumRead);
}

static int
TR2fread_sprite_texture(
  tr2_sprite_texture *buf, size_t count, tr2_file *fp)
{
  int NumRead = 0, i;

  for (i = 0; i < count; i++) {
    NumRead += TR2fread1_bitu16(&buf[i].Tile, fp);
    NumRead += TR2fread1_bitu8(&buf[i].x, fp);
    NumRead += TR2fread1_bitu8(&buf[i].y, fp);
    NumRead += TR2fread1_bitu16(&buf[i].Width, fp);
    NumRead += TR2fread1_bitu16(&buf[i].Height, fp);
    NumRead += TR2fread1_bit16(&buf[i].LeftSide, fp);
    NumRead += TR2fread1_bit16(&buf[i].TopSide, fp);
    NumRead += TR2fread1_bit16(&buf[i].RightSide, fp);
    NumRead += TR2fread1_bit16(&buf[i].BottomSide, fp);
  }
  return(NumRead);
}

static int
TR2fread_sprite_sequence(
  tr2_sprite_sequence *buf, size_t count, tr2_file *fp)
{
  int NumRead = 0, i;

  for (i = 0; i < count; i++) {
    NumRead += TR2fread1_bit32(&buf[i].ObjectID, fp);
    NumRead += TR2fread1_bit16(&buf[i].NegativeLength, fp);
    NumRead += TR2fread1_bit16(&buf[i].Offset, fp);
  }
  return(NumRead);
}

static int
TR2fread_item(
  tr2_item *buf, size_t count, tr2_file *fp)
{
  int NumRead = 0, i;

  for (i = 0; i < count; i++) {
    NumRead += TR2fread1_bit16(&buf[i].ObjectID, fp);
    NumRead += TR2fread1_bit16(&buf[i].Room, fp);
    NumRead += TR2fread1_bit32(&buf[i].x, fp);
    NumRead += TR2fread1_bit32(&buf[i].y, fp);
    NumRead += TR2fread1_bit32(&buf[i].z, fp);
    NumRead += TR2fread1_bit16(&buf[i].Angle, fp);
    NumRead += TR2fread1_bit16(&buf[i].Intensity1, fp);
    if (fp->Version == TombRaider_1) buf[i].Intensity2 = buf[i].Intensity1;
    else NumRead += TR2fread1_bit16(&buf[i].Intensity2, fp);
    NumRead += TR2fread1_bit16(&buf[i].Flags, fp);
  }
  return(NumRead);
}

static int
TR2fread_sound_source(
  tr2_sound_source *buf, size_t count, tr2_file *fp)
{
  int NumRead = 0, i;

  for (i = 0; i < count; i++) {
    NumRead += TR2fread1_bit32(&buf[i].x, fp);
    NumRead += TR2fread1_bit32(&buf[i].y, fp);
    NumRead += TR2fread1_bit32(&buf[i].z, fp);
    NumRead += TR2fread1_bitu16(&buf[i].SoundID, fp);
    NumRead += TR2fread1_bitu16(&buf[i].Flags, fp);
    if (fp->Version != TombRaider_4) /*nop*/;
    else NumRead += TR2fread_bitu8(buf[i].junk, dimof(buf[i].junk), fp);
  }
  return(NumRead);
}

static int
TR2fread_box(
  tr2_box *buf, size_t count, tr2_file *fp)
{
  int NumRead = 0, i;
  bit32 Zmin, Zmax, Xmin, Xmax;

  for (i = 0; i < count; i++) {
    if (fp->Version == TombRaider_1) {
      NumRead += TR2fread1_bit32(&Zmin, fp); buf[i].Zmin = Zmin / 1024;
      NumRead += TR2fread1_bit32(&Zmax, fp); buf[i].Zmax = Zmax / 1024;
      NumRead += TR2fread1_bit32(&Xmin, fp); buf[i].Xmin = Xmin / 1024;
      NumRead += TR2fread1_bit32(&Xmax, fp); buf[i].Xmax = Xmax / 1024;
    }
    else {
      NumRead += TR2fread1_bitu8(&buf[i].Zmin, fp);
      NumRead += TR2fread1_bitu8(&buf[i].Zmax, fp);
      NumRead += TR2fread1_bitu8(&buf[i].Xmin, fp);
      NumRead += TR2fread1_bitu8(&buf[i].Xmax, fp);
    }
    NumRead += TR2fread1_bit16(&buf[i].TrueFloor, fp);
    NumRead += TR2fread1_bit16(&buf[i].OverlapIndex, fp);
  }
  return(NumRead);
}

static int
TR2fread_camera(
  tr2_camera *buf, size_t count, tr2_file *fp)
{
  int NumRead = 0, i;

  for (i = 0; i < count; i++) {
    NumRead += TR2fread1_bit32(&buf[i].x, fp);
    NumRead += TR2fread1_bit32(&buf[i].y, fp);
    NumRead += TR2fread1_bit32(&buf[i].z, fp);
    NumRead += TR2fread1_bit16(&buf[i].Room, fp);
    NumRead += TR2fread1_bitu16(&buf[i].Unknown1, fp);
  }
  return(NumRead);
}

static int
TR2fread_sound_details(
  tr2_sound_details *buf, size_t count, tr2_file *fp)
{
  int NumRead = 0, i;

  for (i = 0; i < count; i++) {
    NumRead += TR2fread1_bit16(&buf[i].Sample, fp);
    NumRead += TR2fread1_bit16(&buf[i].Volume, fp);
    NumRead += TR2fread1_bit16(&buf[i].SoundRange, fp);
    NumRead += TR2fread1_bit16(&buf[i].Flags, fp);
  }
  return(NumRead);
}

static int
TR2fread_cinematic_frame(
  tr2_cinematic_frame *buf, size_t count, tr2_file *fp)
{
  int NumRead = 0, i;

  for (i = 0; i < count; i++) {
    NumRead += TR2fread1_bit16(&buf[i].rotY, fp);
    NumRead += TR2fread1_bit16(&buf[i].rotZ, fp);
    NumRead += TR2fread1_bit16(&buf[i].rotZ2, fp);
    NumRead += TR2fread1_bit16(&buf[i].posZ, fp);
    NumRead += TR2fread1_bit16(&buf[i].posY, fp);
    NumRead += TR2fread1_bit16(&buf[i].posX, fp);
    NumRead += TR2fread1_bit16(&buf[i].Unknown1, fp);
    NumRead += TR2fread1_bit16(&buf[i].rotX, fp);
    if (fp->Version != TombRaider_4) /*nop*/;
    else NumRead += TR2fread_bitu8(buf[i].filler, dimof(buf[i].filler), fp);
  }
  return(NumRead);
}

/*
 * TR2fread_Room -- read a room
 */

static void
TR2fread_RoomData(
   tr2_room_data *Data, tr2_file *fp)
{
   bitu32 DataSize;

   /* identify vertices */
   TR2fread1_bit16(&Data->NumVertices, fp);
   DataSize = Data->NumVertices * sizeof(tr2_vertex_room);
   if (Data->NumVertices > 0) {
      Data->Vertices = (tr2_vertex_room *)TR2alloc(DataSize);
      TR2fread_vertex_room(Data->Vertices, Data->NumVertices, fp);
      }

   /* identify rectangles */
   TR2fread1_bit16(&Data->NumRectangles, fp);
   DataSize = Data->NumRectangles * sizeof(tr2_face4);
   if (Data->NumRectangles > 0) {
      Data->Rectangles = (tr2_face4 *)TR2alloc(DataSize);
      TR2fread_face4(Data->Rectangles, Data->NumRectangles, fp);
      }

   /* identify triangles */
   TR2fread1_bit16(&Data->NumTriangles, fp);
   DataSize = Data->NumTriangles * sizeof(tr2_face3);
   if (Data->NumTriangles > 0) {
      Data->Triangles = (tr2_face3 *)TR2alloc(DataSize);
      TR2fread_face3(Data->Triangles, Data->NumTriangles, fp);
      }

   /* identify sprites */
   TR2fread1_bit16(&Data->NumSprites, fp);
   DataSize = Data->NumSprites * sizeof(tr2_room_sprite);
   if (Data->NumSprites > 0) {
      Data->Sprites = (tr2_room_sprite *)TR2alloc(DataSize);
      TR2fread_room_sprite(Data->Sprites, Data->NumSprites, fp);
      }

   return;
}

static void
TR2fread_Room(
   tr2_room *Room, tr2_file *fp)
{
   tr2_file  tb, *bp = &tb;

   /* read RoomInfo */
   TR2fread_room_info(&Room->info, 1, fp);

   /* read raw data for rest of room */
   TR2fread1_bitu32(&Room->NumDataWords, fp);
   Room->Data = (bitu8 *)TR2alloc(sizeof(bitu16) * Room->NumDataWords);
   TR2fread(Room->Data, sizeof(bitu16), Room->NumDataWords, fp);

   /* read room data */
   TR2fmake(bp, fp->Version, NULL, Room->Data, sizeof(bitu16) * Room->NumDataWords);
   TR2fread_RoomData(&Room->RoomData, bp);

   /* free the raw room data */
   free(Room->Data);
   Room->Data = NULL;  // mark it so we don't accidentally free it again later

   /* read door info */
   TR2fread1_bitu16(&Room->NumPortals, fp);
   Room->Portals = (struct tr2_room_portal_struct *)TR2alloc(sizeof(tr2_room_portal) * Room->NumPortals);
   TR2fread_room_portal(Room->Portals, Room->NumPortals, fp);
      
   /* read sector info */
   TR2fread1_bitu16(&Room->NumZsectors, fp);
   TR2fread1_bitu16(&Room->NumXsectors, fp);
   Room->SectorList = (struct tr2_room_sector_struct *)TR2alloc(sizeof(tr2_room_sector) * Room->NumZsectors * Room->NumXsectors);
   TR2fread_room_sector(Room->SectorList, Room->NumZsectors * Room->NumXsectors, fp);

   /* read room lighting & mode */
   TR2fread1_bit16(&Room->Intensity1, fp);  // Is this intensity or LightMode?
   if (fp->Version == TombRaider_1) Room->Intensity2 = Room->Intensity1;
   else TR2fread1_bit16(&Room->Intensity2, fp);
   if (fp->Version != TombRaider_2) Room->LightMode = 0;  // Fake TR2 record
   else TR2fread1_bit16(&Room->LightMode, fp);

   /* read room lighting info */
   TR2fread1_bitu16(&Room->NumLights, fp);
   if (Room->NumLights > 0) {
      Room->Lights = (struct tr2_room_light_struct *)TR2alloc(sizeof(tr2_room_light) * Room->NumLights);
      TR2fread_room_light(Room->Lights, Room->NumLights, fp);
      }
 
   /* read Static Mesh Data */
   TR2fread1_bitu16(&Room->NumStaticMeshes, fp);
   if (Room->NumStaticMeshes > 0) {
      Room->StaticMeshes = (tr2_room_staticmesh *)TR2alloc(Room->NumStaticMeshes * sizeof(tr2_room_staticmesh));
      TR2fread_room_staticmesh(Room->StaticMeshes, Room->NumStaticMeshes, fp);
      }

   /* read AlternateRoom, Flags */
   TR2fread1_bit16(&Room->AlternateRoom, fp);
   TR2fread1_bit16(&Room->Flags, fp);

   /* read TR3 room light colour */
   if (fp->Version >= TombRaider_3) {
      TR2fread_colour(&Room->RoomLightColour, 1, fp);
      }

   return;
}

/*
 * TR2fread_Mesh -- read a mesh
 */

static int
TR2fread_Mesh(
   tr2_mesh *Mesh, tr2_file *fp)
{
   int    NumRead;
   bitu32 size, i;
   int    NegativeSize;
   bitu8  pad[4] = {0,0,0,0};

   /* get Centre + CollisionSize */
   NumRead += TR2fread_vertex(&Mesh->Centre, 1, fp);
   NumRead += TR2fread1_bit32(&Mesh->CollisionSize, fp);

   /* get number of vertices */
   NumRead += TR2fread1_bit16(&Mesh->NumVertices, fp);
   Mesh->NumVertices = (bit16)abs(Mesh->NumVertices);

   /* get vertex list */
   size = sizeof(tr2_vertex) * Mesh->NumVertices;
   Mesh->Vertices = (tr2_vertex *)TR2alloc(size);
   NumRead += TR2fread_vertex(Mesh->Vertices, Mesh->NumVertices, fp);

   /* get number of normals */
   NumRead += TR2fread1_bit16(&Mesh->NumNormals, fp);
   NegativeSize = (Mesh->NumNormals < 0);
   Mesh->NumNormals = (bit16)abs(Mesh->NumNormals);

   /* get normal list */
   if (NegativeSize) {
      NegativeSize = 0;
      size = Mesh->NumNormals * sizeof(bitu16);
      Mesh->MeshLights = (bit16 *)TR2alloc(size);
      NumRead += TR2fread_bit16(Mesh->MeshLights, Mesh->NumNormals, fp);
      }
   else {
      size = sizeof(tr2_vertex) * Mesh->NumNormals;
      Mesh->Normals = (tr2_vertex *)TR2alloc(size);
      NumRead += TR2fread_vertex(Mesh->Normals, Mesh->NumNormals, fp);
      }

   /* get number of textured rectangles */
   NumRead += TR2fread1_bit16(&Mesh->NumTexturedRectangles, fp);
   Mesh->NumTexturedRectangles = (bit16)abs(Mesh->NumTexturedRectangles);

   /* get list of textured rectangles */
   size = sizeof(tr2_face4) * Mesh->NumTexturedRectangles;
   Mesh->TexturedRectangles = (tr2_face4 *)TR2alloc(size);
   if (Mesh->NumTexturedRectangles > 0) {
      bitu16 junk;
      for (i = 0; i < Mesh->NumTexturedRectangles; i++) {
         NumRead += TR2fread_face4(&Mesh->TexturedRectangles[i], 1, fp);
         if (fp->Version != TombRaider_4) /*nop*/;
         else NumRead += TR2fread1_bitu16(&junk, fp);
         }
      }

   /* get number of textured triangles */
   NumRead += TR2fread1_bit16(&Mesh->NumTexturedTriangles, fp);
   Mesh->NumTexturedTriangles = (bit16)abs(Mesh->NumTexturedTriangles);

   /* get list of textured triangles */
   size = sizeof(tr2_face3) * Mesh->NumTexturedTriangles;
   Mesh->TexturedTriangles = (tr2_face3 *)TR2alloc(size);
   if (Mesh->NumTexturedTriangles > 0) {
      bitu16 junk;
      for (i = 0; i < Mesh->NumTexturedTriangles; i++) {
         NumRead += TR2fread_face3(&Mesh->TexturedTriangles[i], 1, fp);
         if (fp->Version != TombRaider_4) /*nop*/;
         else NumRead += TR2fread1_bitu16(&junk, fp);
         }
      }

   if (fp->Version == TombRaider_4) {  ////////////////////////////////
      Mesh->NumColouredRectangles = 0;
      Mesh->NumColouredTriangles = 0;
      return(NumRead);
      }

   /* get number of coloured rectangles */
   NumRead += TR2fread1_bit16(&Mesh->NumColouredRectangles, fp);
   Mesh->NumColouredRectangles = (bit16)abs(Mesh->NumColouredRectangles);

   /* get list of coloured rectangles */
   size = sizeof(tr2_face4) * Mesh->NumColouredRectangles;
   Mesh->ColouredRectangles = (tr2_face4 *)TR2alloc(size);
   if (Mesh->NumColouredRectangles > 0) {
      NumRead += TR2fread_face4c(Mesh->ColouredRectangles, Mesh->NumColouredRectangles, fp);
      }

   /* get number of coloured triangles */
   NumRead += TR2fread1_bit16(&Mesh->NumColouredTriangles, fp);
   Mesh->NumColouredTriangles = (bit16)abs(Mesh->NumColouredTriangles);

   /* get list of coloured triangles */
   size = sizeof(tr2_face3) * Mesh->NumColouredTriangles;
   Mesh->ColouredTriangles = (tr2_face3 *)TR2alloc(size);
   if (Mesh->NumColouredTriangles > 0) {
      NumRead += TR2fread_face3c(Mesh->ColouredTriangles, Mesh->NumColouredTriangles, fp);
      }

   /* align to 4 byte boundary */
   if (NumRead & 0x03) {
      NumRead += TR2fread_bitu8(pad, 4 - (NumRead & 0x03), fp);
      }

   return(NumRead);
}

/*
 * TR2fread_Level -- read a level
 */

static void
TR2fread_Level(
   TR2_Level *Level, tr2_file *fp)
{
   tr2_file  tb, *bp = &tb;
   int i,
       j;
   bitu32    NumMeshDataWords,
             NumMeshPointers,
             *MeshPointerList;
   bitu8     *RawMeshData;

   /* read the version */
   TR2fread1_bitu32(&Level->Version, fp);
   Level->EngineVersion = TR2getVersion(Level->Version); // TombRaider_1, TombRaider_2, TombRaider_3
   fp->Version = Level->EngineVersion;
   if (Level->EngineVersion == TombRaider_UnknownVersion) {
      fprintf(stderr, "Sorry, unknown file format 0x%08lx.\n", Level->Version);
      exit(2);
      }
   DUMPRD(fp, NULL);

   if (Level->EngineVersion == TombRaider_4) {
#ifdef INCLUDE_TR4_SUPPORT
      bitu16 tmp16;
      bitu32 CompressedSize,
             UncompressedSize;
      bitu8  *CompressedData,
             *UncompressedData;
      int    zerr;

      // read three unknown bitu16s
      TR2fread1_bitu16(&tmp16, fp);
      TR2fread1_bitu16(&tmp16, fp);
      TR2fread1_bitu16(&tmp16, fp);

      // read the sizes of the 32-bit textures
      TR2fread1_bitu32(&UncompressedSize, fp);
      TR2fread1_bitu32(&CompressedSize, fp);
      // skip the 32-bit textures
      TR2fseek(fp, CompressedSize, SEEK_CUR);

      // read the sizes of the 16-bit textures, set NumTextiles
      TR2fread1_bitu32(&UncompressedSize, fp);
      TR2fread1_bitu32(&CompressedSize, fp); 
      Level->NumTextiles = UncompressedSize / (2*256*256);

      // allocate temporary buffers for decompression
      CompressedData = (bitu8 *)TR2alloc(CompressedSize);
      UncompressedData = (bitu8 *)TR2alloc(UncompressedSize);

      // decompress the textures
      TR2fread(CompressedData, CompressedSize, 1, fp);
      zerr = uncompress(UncompressedData, &UncompressedSize, CompressedData, CompressedSize);
      TR2fmake(bp, Level->EngineVersion, NULL, UncompressedData, UncompressedSize);

      // read the 16-bit textures
      Level->Textile16 = (struct tr2_textile16_struct *)TR2alloc(sizeof(tr2_textile16) * Level->NumTextiles);
      TR2fread_textile16(Level->Textile16, Level->NumTextiles, bp);

      // free the temporary buffers
      free((char *)CompressedData);
      free((char *)UncompressedData);

      // read the sizes of the remaining textures
      TR2fread1_bitu32(&UncompressedSize, fp);
      TR2fread1_bitu32(&CompressedSize, fp);
      // skip the remaining textures
      TR2fseek(fp, CompressedSize, SEEK_CUR);

      // read the sizes of the level data
      TR2fread1_bitu32(&UncompressedSize, fp);
      TR2fread1_bitu32(&CompressedSize, fp);

      // allocate temporary buffers for decompression
      CompressedData = (bitu8 *)TR2alloc(CompressedSize);
      UncompressedData = (bitu8 *)TR2alloc(UncompressedSize);

      // decompress the level data
      TR2fread(CompressedData, CompressedSize, 1, fp);
      zerr = uncompress(UncompressedData, &UncompressedSize, CompressedData, CompressedSize);

      // force TR2fread() to read the decompressed data
      TR2fmake(fp, Level->EngineVersion, NULL, UncompressedData, UncompressedSize);

#else    // !INCLUDE_TR4_SUPPORT
      fprintf(stderr, "Sorry, TR4 support was not compiled into this viewer.\n");
      exit(2);
#endif   // INCLUDE_TR4_SUPPORT
      }

   if (Level->EngineVersion == TombRaider_2 || Level->EngineVersion == TombRaider_3) {
      /* read the 8-bit palette */
      TR2fread_colour(Level->Palette8, dimof(Level->Palette8), fp);
      PRINTF(("read Palette8[%d]\n", dimof(Level->Palette8)));
      DUMPRD(fp, "Palette8");

      /* read 16-bit palette */
      TR2fread_bitu32(Level->Palette16, dimof(Level->Palette16), fp);
      PRINTF(("read Palette16[%d]\n", dimof(Level->Palette16)));
      DUMPRD(fp, "Palette16");
      }

   if (Level->EngineVersion != TombRaider_4) {
      /* read the textiles */
      TR2fread1_bitu32(&Level->NumTextiles, fp);
      /* 8-bit textiles come first */
      Level->Textile8 = (struct tr2_textile8_struct *)TR2alloc(sizeof(tr2_textile8) * Level->NumTextiles);
      TR2fread_textile8(Level->Textile8, Level->NumTextiles, fp);
      PRINTF(("read %lu Textile8\n", Level->NumTextiles));
      DUMPRD(fp, "Textile8");
      /* 16-bit textiles come second */
      Level->Textile16 = (struct tr2_textile16_struct *)TR2alloc(sizeof(tr2_textile16) * Level->NumTextiles);
      if (Level->EngineVersion != TombRaider_1) {
         TR2fread_textile16(Level->Textile16, Level->NumTextiles, fp);
         PRINTF(("read %lu Textile16\n", Level->NumTextiles));
         DUMPRD(fp, "Textile16");
         }
      }

   /* 32-bit unknown - seems to always be 0 */
   TR2fread1_bitu32(&Level->UnknownT, fp);

   /* read rooms */
   TR2fread1_bitu16(&Level->NumRooms, fp);
   Level->Rooms = (struct tr2_room_struct *)TR2alloc(sizeof(tr2_room) * Level->NumRooms);
   for (i = 0; i < Level->NumRooms; ++i) {
      TR2fread_Room(&Level->Rooms[i], fp);
      }
   PRINTF(("read %u Rooms\n", Level->NumRooms));
   DUMPRD(fp, "Rooms");
  
   /* read floor data */
   /*
    * Really, FloorData should be a per-sector dynamic allocation;  however,
    * that requires a parser that can accurately determine where one sector's
    * FloorData ends and another's begins.  Until we have that, we'll stick to
    * this crude (but effective) method...
    */

   TR2fread1_bitu32(&Level->NumFloorData, fp);
   if (Level->NumFloorData > 0) {
      Level->FloorData = (bitu16 *)TR2alloc(sizeof(bitu16) * Level->NumFloorData);
      TR2fread_bitu16(Level->FloorData, Level->NumFloorData, fp);
      }
   PRINTF(("read %lu FloorData\n", Level->NumFloorData));
   DUMPRD(fp, "FloorData");
  
   /* read raw mesh data */
   TR2fread1_bitu32(&NumMeshDataWords, fp);
   DUMPRD(fp, "MeshSize");
   RawMeshData = (bitu8 *)TR2alloc(sizeof(bitu16) * NumMeshDataWords);
   TR2fread(RawMeshData, sizeof(bitu16), NumMeshDataWords, fp);
   TR2fmake(bp, fp->Version, NULL, RawMeshData, sizeof(bitu16) * NumMeshDataWords);

   /* read mesh pointers */
   TR2fread1_bitu32(&NumMeshPointers, fp);
   MeshPointerList = (bitu32 *)TR2alloc(sizeof(bitu32) * NumMeshPointers);
   TR2fread_bitu32(MeshPointerList, NumMeshPointers, fp);
   PRINTF(("read %lu MeshPointers\n", NumMeshPointers));

   /* alloc space for meshes */
   Level->NumMeshes = NumMeshPointers;
   Level->Meshes = (struct tr2_mesh_struct *)TR2alloc(sizeof(tr2_mesh) * Level->NumMeshes);

   /* extract meshes */
   for (i = 0; i < NumMeshPointers; i++) {
      TR2fseek(bp, MeshPointerList[i], SEEK_SET);
      TR2fread_Mesh(&Level->Meshes[i], bp);
      SAVERD(bp, MeshPointerList[i], NULL);
      }
   SAVERD(NULL, 0, "Meshes");
   TR2fseek(bp, 0, SEEK_END);
   DUMPRD(bp, NULL);
   free(RawMeshData);
   free(MeshPointerList);
   PRINTF(("read %u Meshes\n", Level->NumMeshes));
   DUMPRD(fp, "MeshPointers");

   /* read animations */
   TR2fread1_bitu32(&Level->NumAnimations, fp);
   if (Level->NumAnimations > 0) {
      Level->Animations = (struct tr2_animation_struct *)TR2alloc(sizeof(tr2_animation) * Level->NumAnimations);
      TR2fread_animation(Level->Animations, Level->NumAnimations, fp);
      }
   PRINTF(("read %lu Animations\n", Level->NumAnimations));
   DUMPRD(fp, "Animations");
  
   /* read state changes */
   TR2fread1_bitu32(&Level->NumStateChanges, fp);
   if (Level->NumStateChanges > 0) {
      Level->StateChanges = (struct tr2_state_change_struct *)TR2alloc(sizeof(tr2_state_change) * Level->NumStateChanges);
      TR2fread_state_change(Level->StateChanges, Level->NumStateChanges, fp);
      }
   PRINTF(("read %lu StateChanges\n", Level->NumStateChanges));
   DUMPRD(fp, "StateChanges");
  
   /* read AnimDispatches */
   TR2fread1_bitu32(&Level->NumAnimDispatches, fp);
   if (Level->NumAnimDispatches > 0) {
      Level->AnimDispatches = (struct tr2_anim_dispatch_struct *)TR2alloc(sizeof(tr2_anim_dispatch) * Level->NumAnimDispatches);
      TR2fread_anim_dispatch(Level->AnimDispatches, Level->NumAnimDispatches, fp);
      }
   PRINTF(("read %lu AnimDispatches\n", Level->NumAnimDispatches));
   DUMPRD(fp, "AnimDispatches");
  
   /* read anim commands */
   TR2fread1_bitu32(&Level->NumAnimCommands, fp);
   if (Level->NumAnimCommands > 0) {
      Level->AnimCommands = (struct tr2_anim_command_struct *)TR2alloc(sizeof(tr2_anim_command) * Level->NumAnimCommands);
      TR2fread_anim_command(Level->AnimCommands, Level->NumAnimCommands, fp);
      }
   PRINTF(("read %lu AnimCommands\n", Level->NumAnimCommands));
   DUMPRD(fp, "AnimCommands");

   /* read MeshTrees */
   TR2fread1_bitu32(&Level->NumMeshTrees, fp);
   if (Level->NumMeshTrees > 0) {
      Level->MeshTrees = (struct tr2_meshtree_struct *)TR2alloc(sizeof(bit32) * Level->NumMeshTrees);
      TR2fread_meshtree(Level->MeshTrees, Level->NumMeshTrees / (sizeof(tr2_meshtree)/sizeof(bit32)), fp);
      }
   PRINTF(("read %lu MeshTrees (%lu/%u)\n", Level->NumMeshTrees / (sizeof(tr2_meshtree)/sizeof(bit32)),
   Level->NumMeshTrees, (sizeof(tr2_meshtree)/sizeof(bit32))));
   DUMPRD(fp, "MeshTrees");
  
   /* read frames */
   TR2fread1_bitu32(&Level->NumFrames, fp);
   if (Level->NumFrames > 0) {
      Level->Frames = (bitu16 *)TR2alloc(sizeof(bitu16) * Level->NumFrames);
      TR2fread_bitu16(Level->Frames, Level->NumFrames, fp);
      if (Level->EngineVersion == TombRaider_1) {
         ioConvertTR1Frames(Level);
         }
      }
   PRINTF(("read %lu Frames\n", Level->NumFrames));
   DUMPRD(fp, "Frames");

   /* read moveables */
   TR2fread1_bitu32(&Level->NumMoveables, fp);
   if (Level->NumMoveables > 0) {
      Level->Moveables = (struct tr2_moveable_struct *)TR2alloc(sizeof(tr2_moveable) * Level->NumMoveables);
      TR2fread_moveable(Level->Moveables, Level->NumMoveables, fp);
      }
   PRINTF(("read %lu Moveables\n", Level->NumMoveables));
   DUMPRD(fp, "Moveables");

   /* read static meshes */
   TR2fread1_bitu32(&Level->NumStaticMeshes, fp);
   if (Level->NumStaticMeshes > 0) {
      Level->StaticMeshes = (tr2_staticmesh *)TR2alloc(Level->NumStaticMeshes * sizeof(tr2_staticmesh));
      TR2fread_staticmesh(Level->StaticMeshes, Level->NumStaticMeshes, fp);
      }
   PRINTF(("read %lu StaticMeshes\n", Level->NumStaticMeshes));
   DUMPRD(fp, "StaticMeshes");

   if (Level->EngineVersion < TombRaider_3) {  
      /* read object textures */
      TR2fread1_bitu32(&Level->NumObjectTextures, fp);
      if (Level->NumObjectTextures > 0) {
         Level->ObjectTextures = (struct tr2_object_texture_struct *)TR2alloc(sizeof(tr2_object_texture) * Level->NumObjectTextures);
         TR2fread_object_texture(Level->ObjectTextures, Level->NumObjectTextures, fp);
         }
      PRINTF(("read %lu ObjectTextures\n", Level->NumObjectTextures));
      DUMPRD(fp, "ObjectTextures");
      }

   if (Level->EngineVersion == TombRaider_4) {
      bitu8 zzbuf[4];
      TR2fread_bitu8(zzbuf, 3, fp); // skip "SPR"
      }

   /* read sprite textures */
   TR2fread1_bitu32(&Level->NumSpriteTextures, fp);
   if (Level->NumSpriteTextures > 0) {
      Level->SpriteTextures = (struct tr2_sprite_texture_struct *)TR2alloc(sizeof(tr2_sprite_texture) * Level->NumSpriteTextures);
      TR2fread_sprite_texture(Level->SpriteTextures,  Level->NumSpriteTextures, fp);
      }
   PRINTF(("read %ld SpriteTextures\n", Level->NumSpriteTextures));
   DUMPRD(fp, "SpriteTextures");

   /* read sprite sequences */
   TR2fread1_bitu32(&Level->NumSpriteSequences, fp);
   if (Level->NumSpriteSequences > 0) {
      Level->SpriteSequences = (struct tr2_sprite_sequence_struct *)TR2alloc(Level->NumSpriteSequences * sizeof(tr2_sprite_sequence));
      TR2fread_sprite_sequence(Level->SpriteSequences, Level->NumSpriteSequences, fp);
      }
   PRINTF(("read %ld SpriteSequences\n", Level->NumSpriteSequences));
   DUMPRD(fp, "SpriteSequences");

   /* read cameras */
   TR2fread1_bit32(&Level->NumCameras, fp);
   if (Level->NumCameras > 0) {
      Level->Cameras = (tr2_camera *)TR2alloc(sizeof(tr2_camera) * Level->NumCameras);
      TR2fread_camera(Level->Cameras, Level->NumCameras, fp);
      }
   PRINTF(("read %ld Cameras\n", Level->NumCameras));
   DUMPRD(fp, "Cameras");

   /* read sound effects (?) */
   TR2fread1_bit32(&Level->NumSoundSources, fp);
   if (Level->NumSoundSources > 0) {
      Level->SoundSources = (tr2_sound_source *)TR2alloc(sizeof(tr2_sound_source) * Level->NumSoundSources);
      TR2fread_sound_source(Level->SoundSources, Level->NumSoundSources, fp);
      // literally, junk[24] should not be in tr2_sound_source; 
      // instead, do TR2fseek(fp, Level->NumSoundSources * 24, SEEK_CUR) afterwards
      }
   PRINTF(("read %ld SoundSources\n", Level->NumSoundSources));
   DUMPRD(fp, "SoundSources");

   if (Level->EngineVersion == TombRaider_4) {  ////////////////////////////////////////
      bitu32 NumZZ;
      bitu8  zzbuf[16];
      TR2fread1_bitu32(&NumZZ, fp);
      while (NumZZ--) {
         TR2fread_bitu8(zzbuf, dimof(zzbuf), fp);
         }
      }

   /* read boxes */
   TR2fread1_bit32(&Level->NumBoxes, fp);
   if (Level->NumBoxes > 0) {
      Level->Boxes = (tr2_box *)TR2alloc(sizeof(tr2_box) * Level->NumBoxes);
      TR2fread_box(Level->Boxes, Level->NumBoxes, fp);
      }
   PRINTF(("read %ld Boxes\n", Level->NumBoxes));
   DUMPRD(fp, "Boxes");

   /* read overlaps (?) */
   TR2fread1_bit32(&Level->NumOverlaps, fp);
   if (Level->NumOverlaps > 0) {
      Level->Overlaps = (bit16 *)TR2alloc(sizeof(bitu16) * Level->NumOverlaps);
      TR2fread_bit16(Level->Overlaps, Level->NumOverlaps, fp);
      }
   PRINTF(("read %ld Overlaps\n", Level->NumOverlaps));
   DUMPRD(fp, "Overlaps");

   /* read Zones */
   if (Level->NumBoxes > 0) {
      Level->Zones = (bit16 *)TR2alloc(20 * Level->NumBoxes);
      if (Level->EngineVersion == TombRaider_1) {
         TR2fread_bit16(Level->Zones, 6 * Level->NumBoxes, fp);
         }
      else {
         TR2fread_bit16(Level->Zones, 10 * Level->NumBoxes, fp);
         }
      }
   PRINTF(("read %ld Zones\n", ((Level->EngineVersion == TombRaider_1)? 6: 10) * Level->NumBoxes));
   DUMPRD(fp, "Zones");

   /* read animation textures (?) */
   TR2fread1_bit32(&Level->NumAnimatedTextures, fp);
   if (Level->NumAnimatedTextures > 0) {
      Level->AnimatedTextures = (bit16 *)TR2alloc(sizeof(bit16) * Level->NumAnimatedTextures);
      TR2fread_bit16(Level->AnimatedTextures, Level->NumAnimatedTextures, fp);
      }
   PRINTF(("read %ld AnimatedTextures\n", Level->NumAnimatedTextures));
   DUMPRD(fp, "AnimatedTextures");

   if (Level->EngineVersion == TombRaider_4) {
      bitu8 zzbuf[4];
      //!! this should be 3, but we have a bug... 
      TR2fread_bitu8(zzbuf, 4, fp); // skip "TEX"
      }

   if (Level->EngineVersion >= TombRaider_3) {
      /* read object textures */
      TR2fread1_bitu32(&Level->NumObjectTextures, fp);
      if (Level->NumObjectTextures > 0) {
           Level->ObjectTextures = (struct tr2_object_texture_struct *)TR2alloc(sizeof(tr2_object_texture) * Level->NumObjectTextures);
           TR2fread_object_texture(Level->ObjectTextures, Level->NumObjectTextures, fp);
         }
      PRINTF(("read %ld ObjectTextures\n", Level->NumObjectTextures));
      DUMPRD(fp, "ObjectTextures");
      }

   /* read items */
   TR2fread1_bit32(&Level->NumItems, fp);
   if (Level->NumItems > 0) {
      Level->Items = (struct tr2_item_struct *)TR2alloc(sizeof(tr2_item) * Level->NumItems);
      TR2fread_item(Level->Items, Level->NumItems, fp);
      }
   PRINTF(("read %ld Items\n", Level->NumItems));
   DUMPRD(fp, "Items");

   /* read LightMap */
   Level->LightMap = (bitu8 *)TR2alloc(32 * 256);
   if (Level->EngineVersion != TombRaider_4) {
      TR2fread_bitu8(Level->LightMap, 32 * 256, fp);
      PRINTF(("read LightMap[%u]\n", 32 * 256));
      DUMPRD(fp, "LightMap");
      }

   if (Level->EngineVersion == TombRaider_1) {
      /* read the 8-bit palette */
      TR2fread_colour(Level->Palette8, dimof(Level->Palette8), fp);
      PRINTF(("read Palette8[%u]\n", dimof(Level->Palette8)));
      DUMPRD(fp, "Palette8");
      /* build 16-bit textiles from 8-bit */
      /* (no better colours, but creates consistent .TR2 file) */
      for (i = 0; i < (int)Level->NumTextiles; ++i) {
         tr2_colour colour;
         bitu16 tmp;
         for (j = 0; j < (256 * 256); ++j) {
            colour = Level->Palette8[Level->Textile8[i].Tile[j]];
            tmp = (((colour.Red & 0x3f) * 31) / 63) << 10;
            tmp |= (((colour.Green & 0x3f) * 31) / 63) << 5;
            tmp |= (((colour.Blue & 0x3f) * 31) / 63);
            tmp |= ((Level->Textile8[i].Tile[j] != 0) << 15);
            Level->Textile16[i].Tile[j] = tmp;
            }
         }
      /* build 16-bit Palette from 8-bit */
      for (i = 0; i < dimof(Level->Palette8); i++) {
         bitu32 tmp;
         tmp = (((Level->Palette8[i].Blue & 0x3f) * 255) / 63) << 16;
         tmp |= (((Level->Palette8[i].Green & 0x3f) * 255) / 63) << 8;
         tmp |= (((Level->Palette8[i].Red & 0x3f) * 255) / 63);
         Level->Palette16[i] = tmp;
         }
      }

   /* read cinematic frames */
   if (Level->EngineVersion == TombRaider_4) {
      bitu32 NumCinematicFrames; // it's now bitu32
      TR2fread1_bitu32(&NumCinematicFrames, fp);
      Level->NumCinematicFrames = (bitu16)NumCinematicFrames;
      }
   else {
      TR2fread1_bitu16(&Level->NumCinematicFrames, fp);
   }
   if (Level->NumCinematicFrames > 0) {
      Level->CinematicFrames = (tr2_cinematic_frame *)TR2alloc(sizeof(tr2_cinematic_frame) * Level->NumCinematicFrames);
      TR2fread_cinematic_frame(Level->CinematicFrames, Level->NumCinematicFrames, fp);
      }
   PRINTF(("read %d CinematicFrames\n", Level->NumCinematicFrames));
   DUMPRD(fp, "CinematicFrames");

   /* read demodata (?) */
   TR2fread1_bit16(&Level->NumDemoData, fp);
   if (Level->NumDemoData > 0) {
      Level->DemoData = (bitu8 *)TR2alloc(1 * Level->NumDemoData);
      TR2fread_bitu8(Level->DemoData, Level->NumDemoData, fp);
      // There may or may not be endian conversion required here - I have
      // no idea what this data is.
      }
   PRINTF(("read %d DemoData\n", Level->NumDemoData));
   DUMPRD(fp, "DemoData");

   /* read SoundMap */
   Level->SoundMap = (bit16 *)TR2alloc(370 * sizeof(bit16));
   if (Level->EngineVersion == TombRaider_1) {
      memset(Level->SoundMap, 0, 370 * sizeof(bit16));   // KLUDGE!!!
      TR2fread_bit16(Level->SoundMap, 256, fp);
      PRINTF(("read SoundMap[%d]\n", 256));
      DUMPRD(fp, "SoundMap");
      }
   else if (Level->EngineVersion == TombRaider_4) {
      TR2fread_bit16(Level->SoundMap, 370, fp);
      PRINTF(("read SoundMap[%d]\n", 370));
      DUMPRD(fp, "SoundMap");
      }
   else {
      TR2fread_bit16(Level->SoundMap, 370, fp);
      PRINTF(("read SoundMap[%d]\n", 370));
      DUMPRD(fp, "SoundMap");
      }

   /* read SoundDetails */
   TR2fread1_bit32(&Level->NumSoundDetails, fp);
   if (Level->NumSoundDetails > 0) {
      Level->SoundDetails = (struct tr2_sound_details_struct *)TR2alloc(sizeof(tr2_sound_details) * Level->NumSoundDetails);
      TR2fread_sound_details(Level->SoundDetails, Level->NumSoundDetails, fp);
      }
   PRINTF(("read %ld SoundDetails\n", Level->NumSoundDetails));
   DUMPRD(fp, "SoundDetails");

#ifdef NEVER_COMPILE_THIS
   if (Level->EngineVersion == TombRaider_1) {
      bitu32 AIFFsize;
      /* read (skip) sound samples (AIFF) */
      TR2fread1_bitu32(&AIFFsize, fp);
      TR2fseek(fp, AIFFsize, SEEK_CUR);
      }
#endif

   if (Level->EngineVersion != TombRaider_1) {  // KLUDGE!
      /* read sampleindices */
      TR2fread1_bit32(&Level->NumSampleIndices, fp);
      if (Level->NumSampleIndices > 0) { 
         Level->SampleIndices = (bit32 *)TR2alloc(sizeof(bitu32) * Level->NumSampleIndices);
         TR2fread_bit32(Level->SampleIndices, Level->NumSampleIndices, fp);
         }
      PRINTF(("read %ld SampleIndices\n", Level->NumSampleIndices));
      DUMPRD(fp, "SampleIndices");
      }

#ifdef NEVER_COMPILE_THIS
   // convert ItemIDs (for writing a .PHD file as a .TR2 file - doesn't usually work)
   if (Level->EngineVersion == TombRaider_1) {
      for (i = 0; i < (int)Level->NumItems; ++i) {
         Level->Items[i].ObjectID = MapTR1ItemToTR2Item(Level->Items[i].ObjectID);
         }
      for (i = 0; i < (int)Level->NumMoveables; ++i) {
         Level->Moveables[i].ObjectID = MapTR1ItemToTR2Item(Level->Moveables[i].ObjectID);
         }
      for (i = 0; i < (int)Level->NumStaticMeshes; ++i) {
         Level->StaticMeshes[i].ObjectID = MapTR1ItemToTR2Item(Level->StaticMeshes[i].ObjectID);
         }
      for (i = 0; i < (int)Level->NumSpriteSequences; ++i) {
         Level->SpriteSequences[i].ObjectID = MapTR1ItemToTR2Item(Level->SpriteSequences[i].ObjectID);
         }
      }
#endif

   FreeIfUsed(fp->Data);

   return;
}


/*
 * The level reader.  Takes a filename (presumably SOMETHING.TR2),
 * reads and parses the file into dynamically allocated structures.
 * Returns a pointer to the whole parsed level.
 */

TR2_Level *
ReadTR2level(char *FileName)
{
   FILE      *ffp;
   tr2_file  tf, *fp = &tf;
   TR2_Level *Level;

   /* try to open the file */
   if ((ffp = fopen(FileName, READBINARY)) == NULL) {
      perror(FileName);
      return(NULL);
      }
   TR2fmake(fp, TombRaider_2, ffp, NULL, 0);

   /* allocate the base structure */
   Level = (TR2_Level *)TR2alloc(sizeof(TR2_Level));

   /* initialise the filename field */
   /* (unfortunately, strdup() isn't universally available...) */
   Level->FileName = (char *)TR2alloc(strlen(FileName) + 1);
   strcpy(Level->FileName, FileName);

   /* read the level */
   TR2fread_Level(Level, fp);

   /* close the file */
   fclose(ffp);

#ifdef RWTEST
   WriteTR2level(Level, "TestLvl.TR2");
   exit(0);
#endif

   return(Level);
}

