
/*
 * tr2wr.c
 *
 * Tomb Raider II write routines for TRueView
 *
 *** CURRENTLY LACKING:
 ***     Better error control (specifically, any malloc() or fread() error
 ***                           just unceremoniously exits the program)
 ***     Correct handling for unknowns (obviously, I can't correctly
 ***                           handle what I don't understand)
 */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <memory.h>
#include <string.h>
#include <stdlib.h>

#include "config.h"

#ifdef INCLUDE_TR4_SUPPORT
// ZLIB #defines and #include for TR4 compression support
#ifndef LINUX
#define _WINDOWS 1
#define ZLIB_DLL 1
#endif
#include "zlib.h"
#endif // INCLUDE_TR4_SUPPORT

#include "tr2port.h"
#include "tr2.h"

#ifdef RWTEST
#undef WRITEBINARY
#define WRITEBINARY "wb+"
#define DUMPWR(f, s)	ioDumpWr(f, s)
#define SAVEWR(f, b, s)	ioSaveWr(f, b, s)
#else
#define DUMPWR(f, s)
#define SAVEWR(f, b, s)
#endif

/*
 * ioCopyWr -- copy file sections after reading
 */
 
#ifdef RWTEST
static long
ioCopyWr(tr2_file *fp, long offset, char *name)
{
   static int count = 0;
   char fname[64];
   FILE *ffp;
   tr2_file td, *dp = &td;
 
   /* open dump file */
   sprintf(fname, "dumpwr-%03d-%s", count++, name);
   if ((ffp = fopen(fname, "w")) == NULL) {
      perror("ioCopyWr: fopen failed");
      exit(2);
   }
   TR2fmake(dp, TombRaider_2, ffp, NULL, 0);
 
   /* copy stuff */
   offset = TR2fcopy(fp, offset, dp);
 
   /* close dump file */
   if (fclose(ffp) != 0) {
      perror("ioCopyWr: fclose failed");
      exit(2);
   }
 
   return(offset);
}

/*
 * ioDumpWr -- dump file sections after reading
 */
 
static void
ioDumpWr(tr2_file *fp, char *name)
{
   static int filemode = 1;
   static long fileoff = 0, buffoff = 0;
 
   /* initialize */
   if (name == NULL) {
      filemode = (fp->Fp != NULL);
      if (filemode) fileoff = 0;
      else buffoff = fp->DataOffset;
      return;
   }
 
   /* change mode */
   if (filemode && (fp->Fp == NULL)) {
      buffoff = 0;
   }
   if (!filemode && (fp->Fp != NULL)) {
      fileoff += buffoff; buffoff = 0;
   }
   filemode = (fp->Fp != NULL);
 
   /* dump stuff */
   if (filemode) fileoff = ioCopyWr(fp, fileoff, name);
   else          buffoff = ioCopyWr(fp, buffoff, name);
 
   return;
}

/*
 * ioSaveWr -- save file sections after reading
 */
 
static void
ioSaveWr(tr2_file *fp, long offset, char *name)
{
   static tr2_file tb = {0}, *bp = &tb;
 
   if (fp != NULL) {
      TR2fcopy(fp, offset, bp);
   }
 
   if (name != NULL) {
      ioCopyWr(bp, 0, name);
      FreeIfUsed(bp->Data); bp->Data = NULL;
      bp->DataOffset = bp->DataSize = 0;
   }
 
   return;
}
#endif  /* RWTEST */

/*
 * ioLocateDuplicateMesh(int which)
 *
 * Scans Level->Meshes[] for an exact match of Level->Meshes[which]
 * (occurring prior to <which>).  Returns index of match, or -1 if
 * no match was found.
 */
static int ioLocateDuplicateMesh(TR2_Level *Level, int which)
{
   int i,
       size;

   for (i = 0; i < which; ++i) {
      if (Level->Meshes[i].NumVertices == Level->Meshes[which].NumVertices &&
          Level->Meshes[i].NumNormals == Level->Meshes[which].NumNormals &&
          Level->Meshes[i].NumTexturedRectangles == Level->Meshes[which].NumTexturedRectangles &&
          Level->Meshes[i].NumTexturedTriangles == Level->Meshes[which].NumTexturedTriangles &&
          Level->Meshes[i].NumColouredRectangles == Level->Meshes[which].NumColouredRectangles &&
          Level->Meshes[i].NumColouredTriangles == Level->Meshes[which].NumColouredTriangles) {
         /* counts match, see if data matches */
         if (memcmp(&Level->Meshes[i].Centre.x, &Level->Meshes[which].Centre.x, 10))
            return(-1);
         size = Level->Meshes[i].NumVertices * sizeof(tr2_vertex);
         if (memcmp(Level->Meshes[i].Vertices, Level->Meshes[which].Vertices, size))
            return(-1);
         if (Level->Meshes[i].MeshLights != NULL || Level->Meshes[which].MeshLights != NULL) {
            if (Level->Meshes[i].MeshLights == NULL || Level->Meshes[which].MeshLights == NULL)
               return(-1);
            size = Level->Meshes[i].NumNormals * 2;
            if (memcmp(Level->Meshes[i].MeshLights, Level->Meshes[which].MeshLights, size))
               return(-1);
            }
         else {
            if (Level->Meshes[i].Normals == NULL || Level->Meshes[which].Normals == NULL)
               return(-1);
            size = Level->Meshes[i].NumNormals * sizeof(tr2_vertex);
            if (memcmp(Level->Meshes[i].Normals, Level->Meshes[which].Normals, size))
               return(-1);
            }
         size = Level->Meshes[i].NumTexturedRectangles * sizeof(tr2_face4);
         if (size > 0 && memcmp(Level->Meshes[i].TexturedRectangles, Level->Meshes[which].TexturedRectangles, size))
            return(-1);
         size = Level->Meshes[i].NumTexturedTriangles * sizeof(tr2_face3);
         if (size > 0 && memcmp(Level->Meshes[i].TexturedTriangles, Level->Meshes[which].TexturedTriangles, size))
            return(-1);
         size = Level->Meshes[i].NumColouredRectangles * sizeof(tr2_face4);
         if (size > 0 && memcmp(Level->Meshes[i].ColouredRectangles, Level->Meshes[which].ColouredRectangles, size))
            return(-1);
         size = Level->Meshes[i].NumColouredTriangles * sizeof(tr2_face3);
         if (size > 0 && memcmp(Level->Meshes[i].ColouredTriangles, Level->Meshes[which].ColouredTriangles, size))
            return(-1);
         return(i);
         }
      }
   return(-1);
}

/*
 * ioGetMeshDataSize
 */

static size_t
ioGetMeshDataSize(
   tr2_mesh *Mesh)
{
   size_t size = 0;

   size += 10 + (6 * sizeof(bitu16));  // size of unknowns and counts (e.g. NumVertices, NumTexturedTriangles, etc.)
   size += Mesh->NumVertices * sizeof(tr2_vertex);
   if (Mesh->MeshLights != NULL)
      size += Mesh->NumNormals * sizeof(bitu16);
   else
      size += Mesh->NumNormals * sizeof(tr2_vertex);
   size += Mesh->NumTexturedRectangles * sizeof(tr2_face4);
   size += Mesh->NumTexturedTriangles * sizeof(tr2_face3);
   size += Mesh->NumColouredRectangles * sizeof(tr2_face4);
   size += Mesh->NumColouredTriangles * sizeof(tr2_face3);
   size += 8;  // fudge for 4-byte alignment
   return(size);
}

/*
 * ioConstructTR1Frames --
 * construct TR1 Frames[] from TR2 frames
 */

static bitu16 *
ioConstructTR1Frames(
   TR2_Level *Level)
{
   bitu16 *Frames;
   int Base, Offset, Num, i, j;

   if (Level->Frames == NULL)
      return(NULL);

   /* duplicate Frames */
   Frames = (bitu16 *)TR2alloc(sizeof(bitu16) * Level->NumFrames);
   for (i = 0; i < (int)Level->NumFrames; i++) {
      Frames[i] = Level->Frames[i];
      }

   for (Base = 0; Base < (int)Level->NumFrames; Base += 10 + 2*Num) {

      /*
       * Recover Num. If there is some Animation where Anim.FrameOffset
       * corresponds to Base, use Anim.FrameSize. Otherwise, use the
       * Animation with next lower Anim.FrameOffset.
       */

      Offset = -1; Num = 1;
      for (j = 0; j < (int)Level->NumAnimations; ++j) {
         int fo = Level->Animations[j].FrameOffset / 2;
         if ((fo <= Base) && (fo > Offset)) {
            Offset = fo;
            Num = (Level->Animations[j].FrameSize - 10) / 2;
            }
         }

      /*
       * Turn    . . . . . . . . . 2 1 4 3 6 5 6    (Num := 3)
       * into    . . . . . . . . . n 1 2 3 4 5 6    (n = Num)
       */

      for (i = Base+9 + 2*Num-2; i-2 >= Base+9; i-=2) {
         Frames[i] = Level->Frames[i-2];
         }
      Frames[Base+9] = Num;
      }

   return(Frames);
}

/*
 * TR2fwrite_xxx -- write miscellaneous data
 */

static int
TR2fwrite_colour(
  tr2_colour *buf, size_t count, tr2_file *fp)
{
  int NumWrite = 0, i;

  for (i = 0; i < count; i++) {
    NumWrite += TR2fwrite1_bitu8(&buf[i].Red, fp);
    NumWrite += TR2fwrite1_bitu8(&buf[i].Green, fp);
    NumWrite += TR2fwrite1_bitu8(&buf[i].Blue, fp);
  }
  return(NumWrite);
}

static int
TR2fwrite_vertex(
  tr2_vertex *buf, size_t count, tr2_file *fp)
{
  int NumWrite = 0, i;

  for (i = 0; i < count; i++) {
    NumWrite += TR2fwrite1_bit16(&buf[i].x, fp);
    NumWrite += TR2fwrite1_bit16(&buf[i].y, fp);
    NumWrite += TR2fwrite1_bit16(&buf[i].z, fp);
  }
  return(NumWrite);
}

static int
TR2fwrite_face4(
  tr2_face4 *buf, size_t count, tr2_file *fp)
{
  int NumWrite = 0, i;

  for (i = 0; i < count; i++) {
    NumWrite += TR2fwrite_bitu16(buf[i].Vertices, dimof(buf[i].Vertices), fp);
    NumWrite += TR2fwrite1_bitu16(&buf[i].Texture, fp);
  }
  return(NumWrite);
}

static int
TR2fwrite_face4c(
  tr2_face4 *buf, size_t count, tr2_file *fp)
{
  int NumWrite = 0, i;
  bitu16 Texture;

  for (i = 0; i < count; i++) {
    NumWrite += TR2fwrite_bitu16(buf[i].Vertices, dimof(buf[i].Vertices), fp);
    Texture = buf[i].Texture;
    if (fp->Version == TombRaider_1) Texture &= 0xff;
    NumWrite += TR2fwrite1_bitu16(&Texture, fp);
  }
  return(NumWrite);
}

static int
TR2fwrite_face3(
  tr2_face3 *buf, size_t count, tr2_file *fp)
{
  int NumWrite = 0, i;

  for (i = 0; i < count; i++) {
    NumWrite += TR2fwrite_bitu16(buf[i].Vertices, dimof(buf[i].Vertices), fp);
    NumWrite += TR2fwrite1_bitu16(&buf[i].Texture, fp);
  }
  return(NumWrite);
}

static int
TR2fwrite_face3c(
  tr2_face3 *buf, size_t count, tr2_file *fp)
{
  int NumWrite = 0, i;
  bitu16 Texture;

  for (i = 0; i < count; i++) {
    NumWrite += TR2fwrite_bitu16(buf[i].Vertices, dimof(buf[i].Vertices), fp);
    Texture = buf[i].Texture;
    if (fp->Version == TombRaider_1) Texture &= 0xff;
    NumWrite += TR2fwrite1_bitu16(&Texture, fp);
  }
  return(NumWrite);
}

static int
TR2fwrite_textile8(
  tr2_textile8 *buf, size_t count, tr2_file *fp)
{
  int NumWrite = 0, i;

  for (i = 0; i < count; i++) {
    NumWrite += TR2fwrite_bitu8(buf[i].Tile, dimof(buf[i].Tile), fp);
  }
  return(NumWrite);
}

static int
TR2fwrite_textile16(
  tr2_textile16 *buf, size_t count, tr2_file *fp)
{
  int NumWrite = 0, i;

  for (i = 0; i < count; i++) {
    NumWrite += TR2fwrite_bitu16(buf[i].Tile, dimof(buf[i].Tile), fp);
  }
  return(NumWrite);
}

static int
TR2fwrite_room_info(
  tr2_room_info *buf, size_t count, tr2_file *fp)
{
  int NumWrite = 0, i;

  for (i = 0; i < count; i++) {
    NumWrite += TR2fwrite1_bit32(&buf[i].x, fp);
    NumWrite += TR2fwrite1_bit32(&buf[i].z, fp);
    NumWrite += TR2fwrite1_bit32(&buf[i].yBottom, fp);
    NumWrite += TR2fwrite1_bit32(&buf[i].yTop, fp);
  }
  return(NumWrite);
}

static int
TR2fwrite_room_portal(
  tr2_room_portal *buf, size_t count, tr2_file *fp)
{
  int NumWrite = 0, i;

  for (i = 0; i < count; i++) {
    NumWrite += TR2fwrite1_bitu16(&buf[i].AdjoiningRoom, fp);
    NumWrite += TR2fwrite_vertex(&buf[i].Normal, 1, fp);
    NumWrite += TR2fwrite_vertex(buf[i].Vertices, dimof(buf[i].Vertices), fp);
  }
  return(NumWrite);
}

static int
TR2fwrite_room_sector(
  tr2_room_sector *buf, size_t count, tr2_file *fp)
{
  int NumWrite = 0, i;

  for (i = 0; i < count; i++) {
    NumWrite += TR2fwrite1_bitu16(&buf[i].FDindex, fp);
    NumWrite += TR2fwrite1_bitu16(&buf[i].BoxIndex, fp);
    NumWrite += TR2fwrite1_bitu8(&buf[i].RoomBelow, fp);
    NumWrite += TR2fwrite1_bit8(&buf[i].Floor, fp);
    NumWrite += TR2fwrite1_bitu8(&buf[i].RoomAbove, fp);
    NumWrite += TR2fwrite1_bit8(&buf[i].Ceiling, fp);
  }
  return(NumWrite);
}

static int
TR2fwrite_room_light(
  tr2_room_light *buf, size_t count, tr2_file *fp)
{
  int NumWrite = 0, i;

  for (i = 0; i < count; i++) {
    NumWrite += TR2fwrite1_bit32(&buf[i].x, fp);
    NumWrite += TR2fwrite1_bit32(&buf[i].y, fp);
    NumWrite += TR2fwrite1_bit32(&buf[i].z, fp);
    NumWrite += TR2fwrite1_bitu16(&buf[i].Intensity1, fp);
    if (fp->Version == TombRaider_1) /*nop*/;
    else NumWrite += TR2fwrite1_bitu16(&buf[i].Intensity2, fp);
    NumWrite += TR2fwrite1_bitu32(&buf[i].Fade1, fp);
    if (fp->Version == TombRaider_1) /*nop*/;
    else NumWrite += TR2fwrite1_bitu32(&buf[i].Fade2, fp);
    if (fp->Version != TombRaider_4) /*nop*/;
    else NumWrite += TR2fwrite_bitu8(buf[i].junk, dimof(buf[i].junk), fp);
  }
  return(NumWrite);
}

static int
TR2fwrite_vertex_room(
  tr2_vertex_room *buf, size_t count, tr2_file *fp)
{
  int NumWrite = 0, i;

  for (i = 0; i < count; i++) {
    NumWrite += TR2fwrite_vertex(&buf[i].Vertex, 1, fp);
    NumWrite += TR2fwrite1_bit16(&buf[i].Lighting1, fp);
    if (fp->Version == TombRaider_1) /*nop*/;
    else NumWrite += TR2fwrite1_bitu16(&buf[i].Attributes, fp);
    if (fp->Version == TombRaider_1) /*nop*/;
    else NumWrite += TR2fwrite1_bit16(&buf[i].Lighting2, fp);
  }
  return(NumWrite);
}

static int
TR2fwrite_room_sprite(
  tr2_room_sprite *buf, size_t count, tr2_file *fp)
{
  int NumWrite = 0, i;

  for (i = 0; i < count; i++) {
    NumWrite += TR2fwrite1_bit16(&buf[i].Vertex, fp);
    NumWrite += TR2fwrite1_bit16(&buf[i].Texture, fp);
  }
  return(NumWrite);
}

static int
TR2fwrite_room_staticmesh(
  tr2_room_staticmesh *buf, size_t count, tr2_file *fp)
{
  int NumWrite = 0, i;

  for (i = 0; i < count; i++) {
    NumWrite += TR2fwrite1_bit32(&buf[i].x, fp);
    NumWrite += TR2fwrite1_bit32(&buf[i].y, fp);
    NumWrite += TR2fwrite1_bit32(&buf[i].z, fp);
    NumWrite += TR2fwrite1_bitu16(&buf[i].Rotation, fp);
    NumWrite += TR2fwrite1_bitu16(&buf[i].Intensity1, fp);
    if (fp->Version == TombRaider_1) /*nop*/;
    else NumWrite += TR2fwrite1_bitu16(&buf[i].Intensity2, fp);
    NumWrite += TR2fwrite1_bitu16(&buf[i].ObjectID, fp);
  }
  return(NumWrite);
}

static int
TR2fwrite_animation(
  tr2_animation *buf, size_t count, tr2_file *fp)
{
  int NumWrite = 0, i;
  bitu8 FrameSize = 0;

  for (i = 0; i < count; i++) {
    FrameSize = buf[i].FrameSize;
    NumWrite += TR2fwrite1_bitu32(&buf[i].FrameOffset, fp);
    NumWrite += TR2fwrite1_bitu8(&buf[i].FrameRate, fp);
    if (fp->Version == TombRaider_1) buf[i].FrameSize = 0;
    NumWrite += TR2fwrite1_bitu8(&buf[i].FrameSize, fp);
    if (fp->Version == TombRaider_1) buf[i].FrameSize = FrameSize;
    NumWrite += TR2fwrite1_bit16(&buf[i].StateID, fp);
    NumWrite += TR2fwrite1_bit16(&buf[i].Unknown1, fp);
    NumWrite += TR2fwrite1_bit16(&buf[i].Unknown2, fp);
    NumWrite += TR2fwrite1_bit16(&buf[i].Unknown3, fp);
    NumWrite += TR2fwrite1_bit16(&buf[i].Unknown4, fp);
    if (fp->Version != TombRaider_4) /*nop*/;
    else NumWrite += TR2fwrite_bitu8(buf[i].junk, dimof(buf[i].junk), fp);
    NumWrite += TR2fwrite1_bitu16(&buf[i].FrameStart, fp);
    NumWrite += TR2fwrite1_bitu16(&buf[i].FrameEnd, fp);
    NumWrite += TR2fwrite1_bitu16(&buf[i].NextAnimation, fp);
    NumWrite += TR2fwrite1_bitu16(&buf[i].NextFrame, fp);
    NumWrite += TR2fwrite1_bitu16(&buf[i].NumStateChanges, fp);
    NumWrite += TR2fwrite1_bitu16(&buf[i].StateChangeOffset, fp);
    NumWrite += TR2fwrite1_bitu16(&buf[i].NumAnimCommands, fp);
    NumWrite += TR2fwrite1_bitu16(&buf[i].AnimCommand, fp);
  }
  return(NumWrite);
}

static int
TR2fwrite_state_change(
  tr2_state_change *buf, size_t count, tr2_file *fp)
{
  int NumWrite = 0, i;

  for (i = 0; i < count; i++) {
    NumWrite += TR2fwrite1_bitu16(&buf[i].StateID, fp);
    NumWrite += TR2fwrite1_bitu16(&buf[i].NumAnimDispatches, fp);
    NumWrite += TR2fwrite1_bitu16(&buf[i].AnimDispatch, fp);
  }
  return(NumWrite);
}

static int
TR2fwrite_anim_dispatch(
  tr2_anim_dispatch *buf, size_t count, tr2_file *fp)
{
  int NumWrite = 0, i;

  for (i = 0; i < count; i++) {
    NumWrite += TR2fwrite1_bit16(&buf[i].Low, fp);
    NumWrite += TR2fwrite1_bit16(&buf[i].High, fp);
    NumWrite += TR2fwrite1_bit16(&buf[i].NextAnimation, fp);
    NumWrite += TR2fwrite1_bit16(&buf[i].NextFrame, fp);
  }
  return(NumWrite);
}

static int
TR2fwrite_anim_command(
  tr2_anim_command *buf, size_t count, tr2_file *fp)
{
  int NumWrite = 0, i;

  for (i = 0; i < count; i++) {
    NumWrite += TR2fwrite1_bit16(&buf[i].Value, fp);
  }
  return(NumWrite);
}

static int
TR2fwrite_meshtree(
  tr2_meshtree *buf, size_t count, tr2_file *fp)
{
  int NumWrite = 0, i;

  for (i = 0; i < count; i++) {
    NumWrite += TR2fwrite1_bit32(&buf[i].Flags, fp);
    NumWrite += TR2fwrite1_bit32(&buf[i].x, fp);
    NumWrite += TR2fwrite1_bit32(&buf[i].y, fp);
    NumWrite += TR2fwrite1_bit32(&buf[i].z, fp);
  }
  return(NumWrite);
}

static int
TR2fwrite_moveable(
  tr2_moveable *buf, size_t count, tr2_file *fp)
{
  int NumWrite = 0, i;

  for (i = 0; i < count; i++) {
    NumWrite += TR2fwrite1_bitu32(&buf[i].ObjectID, fp);
    NumWrite += TR2fwrite1_bitu16(&buf[i].NumMeshes, fp);
    NumWrite += TR2fwrite1_bitu16(&buf[i].StartingMesh, fp);
    NumWrite += TR2fwrite1_bitu32(&buf[i].MeshTree, fp);
    NumWrite += TR2fwrite1_bitu32(&buf[i].FrameOffset, fp);
    NumWrite += TR2fwrite1_bitu16(&buf[i].Animation, fp);
  }
  return(NumWrite);
}

static int
TR2fwrite_staticmesh(
  tr2_staticmesh *buf, size_t count, tr2_file *fp)
{
  int NumWrite = 0, i;

  for (i = 0; i < count; i++) {
    NumWrite += TR2fwrite1_bitu32(&buf[i].ObjectID, fp);
    NumWrite += TR2fwrite1_bitu16(&buf[i].StartingMesh, fp);
    NumWrite += TR2fwrite_vertex(&buf[i].BoundingBox[0][0], dim2of(buf[i].BoundingBox), fp);
    NumWrite += TR2fwrite1_bitu16(&buf[i].Flags, fp);
  }
  return(NumWrite);
}

static int
TR2fwrite_object_texture_vert(
  tr2_object_texture_vert *buf, size_t count, tr2_file *fp)
{
  int NumWrite = 0, i;

  for (i = 0; i < count; i++) {
    NumWrite += TR2fwrite1_bitu8(&buf[i].Xcoordinate, fp);
    NumWrite += TR2fwrite1_bitu8(&buf[i].Xpixel, fp);
    NumWrite += TR2fwrite1_bitu8(&buf[i].Ycoordinate, fp);
    NumWrite += TR2fwrite1_bitu8(&buf[i].Ypixel, fp);
  }
  return(NumWrite);
}

static int
TR2fwrite_object_texture(
  tr2_object_texture *buf, size_t count, tr2_file *fp)
{
  int NumWrite = 0, i;
  bitu32 Tile;

  for (i = 0; i < count; i++) {
    NumWrite += TR2fwrite1_bitu16(&buf[i].TransparencyFlags, fp);
    if (fp->Version == TombRaider_4) {
      Tile = buf[i].Tile; NumWrite += TR2fwrite1_bitu32(&Tile, fp); }
    else {
      NumWrite += TR2fwrite1_bitu16(&buf[i].Tile, fp); }
    NumWrite += TR2fwrite_object_texture_vert(buf[i].Vertices, dimof(buf[i].Vertices), fp);
    if (fp->Version != TombRaider_4) /*nop*/;
    else NumWrite += TR2fwrite_bitu8(buf[i].filler, dimof(buf[i].filler), fp);
  }
  return(NumWrite);
}

static int
TR2fwrite_sprite_texture(
  tr2_sprite_texture *buf, size_t count, tr2_file *fp)
{
  int NumWrite = 0, i;

  for (i = 0; i < count; i++) {
    NumWrite += TR2fwrite1_bitu16(&buf[i].Tile, fp);
    NumWrite += TR2fwrite1_bitu8(&buf[i].x, fp);
    NumWrite += TR2fwrite1_bitu8(&buf[i].y, fp);
    NumWrite += TR2fwrite1_bitu16(&buf[i].Width, fp);
    NumWrite += TR2fwrite1_bitu16(&buf[i].Height, fp);
    NumWrite += TR2fwrite1_bit16(&buf[i].LeftSide, fp);
    NumWrite += TR2fwrite1_bit16(&buf[i].TopSide, fp);
    NumWrite += TR2fwrite1_bit16(&buf[i].RightSide, fp);
    NumWrite += TR2fwrite1_bit16(&buf[i].BottomSide, fp);
  }
  return(NumWrite);
}

static int
TR2fwrite_sprite_sequence(
  tr2_sprite_sequence *buf, size_t count, tr2_file *fp)
{
  int NumWrite = 0, i;

  for (i = 0; i < count; i++) {
    NumWrite += TR2fwrite1_bit32(&buf[i].ObjectID, fp);
    NumWrite += TR2fwrite1_bit16(&buf[i].NegativeLength, fp);
    NumWrite += TR2fwrite1_bit16(&buf[i].Offset, fp);
  }
  return(NumWrite);
}

static int
TR2fwrite_item(
  tr2_item *buf, size_t count, tr2_file *fp)
{
  int NumWrite = 0, i;

  for (i = 0; i < count; i++) {
    NumWrite += TR2fwrite1_bit16(&buf[i].ObjectID, fp);
    NumWrite += TR2fwrite1_bit16(&buf[i].Room, fp);
    NumWrite += TR2fwrite1_bit32(&buf[i].x, fp);
    NumWrite += TR2fwrite1_bit32(&buf[i].y, fp);
    NumWrite += TR2fwrite1_bit32(&buf[i].z, fp);
    NumWrite += TR2fwrite1_bit16(&buf[i].Angle, fp);
    NumWrite += TR2fwrite1_bit16(&buf[i].Intensity1, fp);
    if (fp->Version == TombRaider_1) /*nop*/;
    else NumWrite += TR2fwrite1_bit16(&buf[i].Intensity2, fp);
    NumWrite += TR2fwrite1_bit16(&buf[i].Flags, fp);
  }
  return(NumWrite);
}

static int
TR2fwrite_sound_source(
  tr2_sound_source *buf, size_t count, tr2_file *fp)
{
  int NumWrite = 0, i;

  for (i = 0; i < count; i++) {
    NumWrite += TR2fwrite1_bit32(&buf[i].x, fp);
    NumWrite += TR2fwrite1_bit32(&buf[i].y, fp);
    NumWrite += TR2fwrite1_bit32(&buf[i].z, fp);
    NumWrite += TR2fwrite1_bitu16(&buf[i].SoundID, fp);
    NumWrite += TR2fwrite1_bitu16(&buf[i].Flags, fp);
    if (fp->Version != TombRaider_4) /*nop*/;
    else NumWrite += TR2fwrite_bitu8(buf[i].junk, dimof(buf[i].junk), fp);
  }
  return(NumWrite);
}

static int
TR2fwrite_box(
  tr2_box *buf, size_t count, tr2_file *fp)
{
  int NumWrite = 0, i;
  bit32 Zmin, Zmax, Xmin, Xmax;

  for (i = 0; i < count; i++) {
    if (fp->Version == TombRaider_1) {
      Zmin = buf[i].Zmin * 1024; Zmax = buf[i].Zmax * 1024 + 1023;
      Xmin = buf[i].Xmin * 1024; Xmax = buf[i].Xmax * 1024 + 1023;
      NumWrite += TR2fwrite1_bit32(&Zmin, fp);
      NumWrite += TR2fwrite1_bit32(&Zmax, fp);
      NumWrite += TR2fwrite1_bit32(&Xmin, fp);
      NumWrite += TR2fwrite1_bit32(&Xmax, fp);
    }
    else {
      NumWrite += TR2fwrite1_bitu8(&buf[i].Zmin, fp);
      NumWrite += TR2fwrite1_bitu8(&buf[i].Zmax, fp);
      NumWrite += TR2fwrite1_bitu8(&buf[i].Xmin, fp);
      NumWrite += TR2fwrite1_bitu8(&buf[i].Xmax, fp);
    }
    NumWrite += TR2fwrite1_bit16(&buf[i].TrueFloor, fp);
    NumWrite += TR2fwrite1_bit16(&buf[i].OverlapIndex, fp);
  }
  return(NumWrite);
}

static int
TR2fwrite_camera(
  tr2_camera *buf, size_t count, tr2_file *fp)
{
  int NumWrite = 0, i;

  for (i = 0; i < count; i++) {
    NumWrite += TR2fwrite1_bit32(&buf[i].x, fp);
    NumWrite += TR2fwrite1_bit32(&buf[i].y, fp);
    NumWrite += TR2fwrite1_bit32(&buf[i].z, fp);
    NumWrite += TR2fwrite1_bit16(&buf[i].Room, fp);
    NumWrite += TR2fwrite1_bitu16(&buf[i].Unknown1, fp);
  }
  return(NumWrite);
}

static int
TR2fwrite_sound_details(
  tr2_sound_details *buf, size_t count, tr2_file *fp)
{
  int NumWrite = 0, i;

  for (i = 0; i < count; i++) {
    NumWrite += TR2fwrite1_bit16(&buf[i].Sample, fp);
    NumWrite += TR2fwrite1_bit16(&buf[i].Volume, fp);
    NumWrite += TR2fwrite1_bit16(&buf[i].SoundRange, fp);
    NumWrite += TR2fwrite1_bit16(&buf[i].Flags, fp);
  }
  return(NumWrite);
}

static int
TR2fwrite_cinematic_frame(
  tr2_cinematic_frame *buf, size_t count, tr2_file *fp)
{
  int NumWrite = 0, i;

  for (i = 0; i < count; i++) {
    NumWrite += TR2fwrite1_bit16(&buf[i].rotY, fp);
    NumWrite += TR2fwrite1_bit16(&buf[i].rotZ, fp);
    NumWrite += TR2fwrite1_bit16(&buf[i].rotZ2, fp);
    NumWrite += TR2fwrite1_bit16(&buf[i].posZ, fp);
    NumWrite += TR2fwrite1_bit16(&buf[i].posY, fp);
    NumWrite += TR2fwrite1_bit16(&buf[i].posX, fp);
    NumWrite += TR2fwrite1_bit16(&buf[i].Unknown1, fp);
    NumWrite += TR2fwrite1_bit16(&buf[i].rotX, fp);
    if (fp->Version != TombRaider_4) /*nop*/;
    else NumWrite += TR2fwrite_bitu8(buf[i].filler, dimof(buf[i].filler), fp);
  }
  return(NumWrite);
}

/*
 * TR2fwrite_Room -- write a room
 */

static int
TR2fwrite_RoomData(
   tr2_room_data *Data, tr2_file *fp)
{
   int NumWrite = 0;

   /* write vertices */
   NumWrite += TR2fwrite1_bit16(&Data->NumVertices, fp);
   if (Data->NumVertices > 0) {
      NumWrite += TR2fwrite_vertex_room(Data->Vertices, Data->NumVertices, fp);
      }

   /* write rectangles */
   NumWrite += TR2fwrite1_bit16(&Data->NumRectangles, fp);
   if (Data->NumRectangles > 0) {
      NumWrite += TR2fwrite_face4(Data->Rectangles, Data->NumRectangles, fp);
      }

   /* write triangles */
   NumWrite += TR2fwrite1_bit16(&Data->NumTriangles, fp);
   if (Data->NumTriangles > 0) {
      NumWrite += TR2fwrite_face3(Data->Triangles, Data->NumTriangles, fp);
      }

   /* write sprites */
   NumWrite += TR2fwrite1_bit16(&Data->NumSprites, fp);
   if (Data->NumSprites > 0) {
      NumWrite += TR2fwrite_room_sprite(Data->Sprites, Data->NumSprites, fp);
      }

   return(NumWrite);
}

static void
TR2fwrite_Room(
   tr2_room *Room, tr2_file *fp)
{
   tr2_file  tb, *bp = &tb;
   size_t    count;
   bitu8     *TempData = NULL;
   bitu32    TempDataSize;
   bitu32    NumDataWords;

   /* write RoomInfo */
   TR2fwrite_room_info(&Room->info, 1, fp);

   /* build raw data for rest of room */
   TempDataSize =
      sizeof(bitu16) + (Room->RoomData.NumVertices * sizeof(tr2_vertex_room)) +
      sizeof(bitu16) + (Room->RoomData.NumRectangles * sizeof(tr2_face4)) +
      sizeof(bitu16) + (Room->RoomData.NumTriangles * sizeof(tr2_face3)) +
      sizeof(bitu16) + (Room->RoomData.NumSprites * sizeof(tr2_room_sprite));
   TempData = (bitu8 *)TR2alloc(TempDataSize); // allocate space for the raw data

   /* copy raw room data to temporary buffer */
   TR2fmake(bp, fp->Version, NULL, TempData, TempDataSize);
   count = (size_t) TR2fwrite_RoomData(&Room->RoomData, bp);

   /* write room data to file */
   NumDataWords = count / sizeof(bitu16);
   TR2fwrite1_bitu32(&NumDataWords, fp);
   TR2fwrite_bitu8(TempData, count, fp);

   /* get rid of our temporary buffer */
   free(TempData);
   TempData = NULL;

   /* write portal info */
   TR2fwrite1_bitu16(&Room->NumPortals, fp); 
   TR2fwrite_room_portal(Room->Portals, Room->NumPortals, fp);

   /* write sector info */
   TR2fwrite1_bitu16(&Room->NumZsectors, fp);
   TR2fwrite1_bitu16(&Room->NumXsectors, fp);
   TR2fwrite_room_sector(Room->SectorList, Room->NumZsectors * Room->NumXsectors, fp);

   /* write Intensity1, Intensity2, LightMode */
   TR2fwrite1_bit16(&Room->Intensity1, fp);
   if (fp->Version == TombRaider_1) /*nop*/;
   else TR2fwrite1_bit16(&Room->Intensity2, fp);
   if (fp->Version != TombRaider_2) /*nop*/;
   else TR2fwrite1_bit16(&Room->LightMode, fp);

   /* write room lighting info */
   TR2fwrite1_bitu16(&Room->NumLights, fp);
   if (Room->NumLights > 0) {
      TR2fwrite_room_light(Room->Lights, Room->NumLights, fp);
      }
 
   /* write Static Mesh Data */
   TR2fwrite1_bitu16(&Room->NumStaticMeshes, fp);
   if (Room->NumStaticMeshes > 0) {
      TR2fwrite_room_staticmesh(Room->StaticMeshes, Room->NumStaticMeshes, fp); 
      }

   /* write AlternateRoom, Flags */
   TR2fwrite1_bit16(&Room->AlternateRoom, fp);
   TR2fwrite1_bit16(&Room->Flags, fp);

   /* write TR3 room light colour */
   if (fp->Version >= TombRaider_3) {
      TR2fwrite_colour(&Room->RoomLightColour, 1, fp);
      }

   return;
}

/*
 * TR2fwrite_Mesh -- write a mesh
 */

static int
TR2fwrite_Mesh(
   tr2_mesh *Mesh, tr2_file *fp)
{
   int NumWrite = 0;
   bitu16 NumNormals;
   bitu8 pad[4] = {0,0,0,0};

   /* write Centre + CollisionSize */
   NumWrite += TR2fwrite_vertex(&Mesh->Centre, 1, fp);
   NumWrite += TR2fwrite1_bit32(&Mesh->CollisionSize, fp);

   /* write Vertices */
   NumWrite += TR2fwrite1_bit16(&Mesh->NumVertices, fp);
   NumWrite += TR2fwrite_vertex(Mesh->Vertices, Mesh->NumVertices, fp);

   /* write number of normals */
   NumNormals = Mesh->NumNormals;
   if (Mesh->MeshLights != NULL)
      NumNormals = -NumNormals;
   NumWrite += TR2fwrite1_bit16(&NumNormals, fp);

   /* write Normals */
   if (Mesh->MeshLights != NULL) {
      NumWrite += TR2fwrite_bit16(Mesh->MeshLights, Mesh->NumNormals, fp);
      }
   else {   // "normal" Normals
      NumWrite += TR2fwrite_vertex(Mesh->Normals, Mesh->NumNormals, fp);
      }

   /* write Textured Rectangles */
   NumWrite += TR2fwrite1_bit16(&Mesh->NumTexturedRectangles, fp);
   if (Mesh->NumTexturedRectangles > 0) {
      NumWrite += TR2fwrite_face4(Mesh->TexturedRectangles, Mesh->NumTexturedRectangles, fp);
      }

   /* write Textured Triangles */
   NumWrite += TR2fwrite1_bit16(&Mesh->NumTexturedTriangles, fp);
   if (Mesh->NumTexturedTriangles > 0) {
      NumWrite += TR2fwrite_face3(Mesh->TexturedTriangles, Mesh->NumTexturedTriangles, fp);
      }

   /* write Coloured Rectangles */
   NumWrite += TR2fwrite1_bit16(&Mesh->NumColouredRectangles, fp);
   if (Mesh->NumColouredRectangles > 0) {
      NumWrite += TR2fwrite_face4c(Mesh->ColouredRectangles, Mesh->NumColouredRectangles, fp);
      }

   /* write Coloured Triangles */
   NumWrite += TR2fwrite1_bit16(&Mesh->NumColouredTriangles, fp);
   if (Mesh->NumColouredTriangles > 0) {
      NumWrite += TR2fwrite_face3c(Mesh->ColouredTriangles, Mesh->NumColouredTriangles, fp);
      }

   /* align to 4 byte boundary */
   if (NumWrite & 0x03) {
      NumWrite += TR2fwrite_bitu8(pad, 4 - (NumWrite & 0x03), fp); 
      }

   return(NumWrite);
}

/*
 * TR2fwrite_Level -- write a level
 */

static void
TR2fwrite_Level(TR2_Level *Level, tr2_file *fp)
{
   tr2_file  tb, *bp = &tb;
   bitu8     *TempData;
   bitu32    TempDataSize,
             MeshDataSize;
   bitu32    NumMeshDataWords,
             NumMeshPointers,
             *MeshPointers;
   bitu32    Version;
   int       i, j;

   /* check version */
   switch (fp->Version) {
      case TombRaider_1: Version = 0x00000020; break;
      case TombRaider_2: Version = 0x0000002d; break;
      case TombRaider_3: Version = 0xff080038; break;
      case TombRaider_4: Version = 0x00345254; break;
      default:           Version = 0x00000000; break;
      }
   /* XXX -- remove this when TR4 supported */
   if (fp->Version == TombRaider_4) {
      Version = 0x00000000;
      fprintf(stderr, "Sorry, can't write TR4 format.\n");
      return;
      }

   /* write the version */
   if (Version == 0x00000000) {
      // temporary hack - always write "TR2 versions"
      fp->Version = TombRaider_2;
      Version = 0x0000002d;
      }
   else if (fp->Version == Level->EngineVersion) {
      Version = Level->Version;
      }
   TR2fwrite1_bitu32(&Version, fp);
   DUMPWR(fp, NULL);

   if (fp->Version == TombRaider_4) {
      /* XXX -- TR4: write compressed stuff */
   }

   if (fp->Version == TombRaider_2 || fp->Version == TombRaider_3) {
      /* write the 8-bit palette */
      TR2fwrite_colour(Level->Palette8, dimof(Level->Palette8), fp);
      PRINTF(("wrote Palette8[%d]\n", dimof(Level->Palette8)));
      DUMPWR(fp, "Palette8");

      /* write 16-bit palette */
      TR2fwrite_bitu32(Level->Palette16, dimof(Level->Palette16), fp);
      PRINTF(("wrote Palette16[%d]\n", dimof(Level->Palette16)));
      DUMPWR(fp, "Palette16");
      }

   if (fp->Version != TombRaider_4) {
      /* write the textiles */
      TR2fwrite1_bitu32(&Level->NumTextiles, fp); 
      /* 8-bit textiles come first */
      TR2fwrite_textile8(Level->Textile8, Level->NumTextiles, fp);
      PRINTF(("wrote %lu Textile8\n", Level->NumTextiles));
      DUMPWR(fp, "Textile8");
      /* 16-bit textiles come second */
      if (fp->Version != TombRaider_1) {
         TR2fwrite_textile16(Level->Textile16, Level->NumTextiles, fp);
         PRINTF(("wrote %lu Textile16\n", Level->NumTextiles));
         DUMPWR(fp, "Textile16");
         }
      }

   /* write 32-bit unknown */
   TR2fwrite1_bitu32(&Level->UnknownT, fp);

   /* write room count */
   TR2fwrite1_bitu16(&Level->NumRooms, fp);

   /* write room details */
   for (i = 0; i < Level->NumRooms; ++i) {
      TR2fwrite_Room(&Level->Rooms[i], fp); 
      }
   PRINTF(("wrote %u Rooms\n", Level->NumRooms));
   DUMPWR(fp, "Rooms");

   /* write floor data */
   /*
    * Really, FloorData should be a per-sector dynamic allocation;  however,
    * that requires a parser that can accurately determine where one sector's
    * FloorData ends and another's begins.  Until we have that, we'll stick to
    * this crude (but effective) method...
    */

   TR2fwrite1_bitu32(&Level->NumFloorData, fp);
   if (Level->NumFloorData > 0) {
      TR2fwrite_bitu16(Level->FloorData, Level->NumFloorData, fp);
      }
   PRINTF(("wrote %lu FloorData\n", Level->NumFloorData));
   DUMPWR(fp, "FloorData");

   /* allocate temporary data buffer for meshes */
   for (i = 0, TempDataSize = 0; i < Level->NumMeshes; i++) {
      TempDataSize += ioGetMeshDataSize(&Level->Meshes[i]);
      }
   TempData = (bitu8 *)TR2alloc(TempDataSize);
   NumMeshPointers = Level->NumMeshes;
   MeshPointers = (bitu32 *)TR2alloc(NumMeshPointers * sizeof(bitu32));
 
   /* generate raw mesh data */
   TR2fmake(bp, fp->Version, NULL, TempData, TempDataSize);
   for (i = 0, MeshDataSize = 0; i < Level->NumMeshes; i++) {
      /* set current mesh pointer */
      if ((j = ioLocateDuplicateMesh(Level, i)) != -1) {
         MeshPointers[i] = MeshPointers[j];
#ifdef RWTEST
         TR2fwrite_Mesh(&Level->Meshes[i], bp);
         SAVEWR(bp, MeshDataSize, NULL);
         TR2fseek(bp, MeshDataSize, SEEK_SET);
#endif
         continue;
         }
      MeshPointers[i] = MeshDataSize;
 
      /* write raw mesh data into buffer */
      MeshDataSize += (bitu32) TR2fwrite_Mesh(&Level->Meshes[i], bp);
      SAVEWR(bp, MeshPointers[i], NULL);
      }
 
   /* write mesh data */
   NumMeshDataWords = MeshDataSize / sizeof(bitu16);
   TR2fwrite1_bitu32(&NumMeshDataWords, fp);
   DUMPWR(fp, "MeshSize");
   SAVEWR(NULL, 0, "Meshes");
   DUMPWR(bp, NULL);
   TR2fwrite_bitu8(TempData, MeshDataSize, fp);
   PRINTF(("wrote %u Meshes\n", Level->NumMeshes));

   /* write mesh pointers */
   TR2fwrite1_bitu32(&NumMeshPointers, fp);
   TR2fwrite_bitu32(MeshPointers, NumMeshPointers, fp);
   PRINTF(("wrote %lu MeshPointers\n", NumMeshPointers));
   DUMPWR(fp, "MeshPointers");
 
   /* free temporary buffers for meshes */
   free(TempData); TempData = NULL;
   free(MeshPointers); MeshPointers = NULL;

   /* write animations */
   TR2fwrite1_bitu32(&Level->NumAnimations, fp);
   if (Level->NumAnimations > 0) {
      TR2fwrite_animation(Level->Animations, Level->NumAnimations, fp);
      }
   PRINTF(("wrote %lu Animations\n", Level->NumAnimations));
   DUMPWR(fp, "Animations");

   /* write state changes */
   TR2fwrite1_bitu32(&Level->NumStateChanges, fp);
   if (Level->NumStateChanges > 0) {
      TR2fwrite_state_change(Level->StateChanges, Level->NumStateChanges, fp);
      }
   PRINTF(("wrote %lu StateChanges\n", Level->NumStateChanges));
   DUMPWR(fp, "StateChanges");

   /* write anim dispatches */
   TR2fwrite1_bitu32(&Level->NumAnimDispatches, fp);
   if (Level->NumAnimDispatches > 0) {
      TR2fwrite_anim_dispatch(Level->AnimDispatches, Level->NumAnimDispatches, fp);
      }
   PRINTF(("wrote %lu AnimDispatches\n", Level->NumAnimDispatches));
   DUMPWR(fp, "AnimDispatches");
  
   /* write AnimCommands */
   TR2fwrite1_bitu32(&Level->NumAnimCommands, fp);
   if (Level->NumAnimCommands > 0) {
      TR2fwrite_anim_command(Level->AnimCommands, Level->NumAnimCommands, fp);
      }
   PRINTF(("wrote %lu AnimCommands\n", Level->NumAnimCommands));
   DUMPWR(fp, "AnimCommands");

   /* write MeshTrees */
   TR2fwrite1_bitu32(&Level->NumMeshTrees, fp);
   if (Level->NumMeshTrees > 0) {
      TR2fwrite_meshtree(Level->MeshTrees, Level->NumMeshTrees / (sizeof(tr2_meshtree)/sizeof(bit32)), fp);
      }
   PRINTF(("wrote %lu MeshTrees\n", Level->NumMeshTrees / (sizeof(tr2_meshtree)/sizeof(bit32))));
   DUMPWR(fp, "MeshTrees");

   /* write frames */
   TR2fwrite1_bitu32(&Level->NumFrames, fp);
   if (Level->NumFrames > 0) {
      bitu16 *Frames = Level->Frames;
      if (fp->Version == TombRaider_1) {
         Frames = ioConstructTR1Frames(Level);
         }
      TR2fwrite_bitu16(Frames, Level->NumFrames, fp);
      }
   PRINTF(("wrote %lu Frames\n", Level->NumFrames));
   DUMPWR(fp, "Frames");

   /* write moveables */
   TR2fwrite1_bitu32(&Level->NumMoveables, fp);
   if (Level->NumMoveables > 0) {
      TR2fwrite_moveable(Level->Moveables, Level->NumMoveables, fp);
      }
   PRINTF(("wrote %lu Moveables\n", Level->NumMoveables));
   DUMPWR(fp, "Moveables");

   /* write static meshes */
   TR2fwrite1_bitu32(&Level->NumStaticMeshes, fp);
   if (Level->NumStaticMeshes > 0) {
      TR2fwrite_staticmesh(Level->StaticMeshes, Level->NumStaticMeshes, fp);
      }
   PRINTF(("wrote %lu StaticMeshes\n", Level->NumStaticMeshes));
   DUMPWR(fp, "StaticMeshes");
  
   if (fp->Version < TombRaider_3) {
      /* write object textures */
      TR2fwrite1_bitu32(&Level->NumObjectTextures, fp);
      if (Level->NumObjectTextures > 0) {
         TR2fwrite_object_texture(Level->ObjectTextures, Level->NumObjectTextures, fp);
         }
      PRINTF(("wrote %lu ObjectTextures\n", Level->NumObjectTextures));
      DUMPWR(fp, "ObjectTextures");
      }

   if (fp->Version == TombRaider_4) {
      /* XXX -- TR4: write "SPR" */
      bitu8 zzbuf[4] = {'S', 'P', 'R', '\0'};
      TR2fwrite_bitu8(zzbuf, 3, fp); // write "SPR"
      }

   /* write sprite textures */
   TR2fwrite1_bitu32(&Level->NumSpriteTextures, fp);
   if (Level->NumSpriteTextures > 0) {
      TR2fwrite_sprite_texture(Level->SpriteTextures, Level->NumSpriteTextures, fp);
      }
   PRINTF(("wrote %ld SpriteTextures\n", Level->NumSpriteTextures));
   DUMPWR(fp, "SpriteTextures");
  
   /* write sprite texture data (?) */
   TR2fwrite1_bitu32(&Level->NumSpriteSequences, fp);
   if (Level->NumSpriteSequences > 0) {
      TR2fwrite_sprite_sequence(Level->SpriteSequences, Level->NumSpriteSequences, fp);
      }
   PRINTF(("wrote %ld SpriteSequences\n", Level->NumSpriteSequences));
   DUMPWR(fp, "SpriteSequences");

   /* write cameras (?) */
   TR2fwrite1_bit32(&Level->NumCameras, fp);
   if (Level->NumCameras > 0) {
      TR2fwrite_camera(Level->Cameras, Level->NumCameras, fp);
      }
   PRINTF(("wrote %ld Cameras\n", Level->NumCameras));
   DUMPWR(fp, "Cameras");

   /* write sound effects */
   TR2fwrite1_bit32(&Level->NumSoundSources, fp);
   if (Level->NumSoundSources > 0) {
      TR2fwrite_sound_source(Level->SoundSources, Level->NumSoundSources, fp);
      }
   PRINTF(("write %ld SoundSources\n", Level->NumSoundSources));
   DUMPWR(fp, "SoundSources");

   if (fp->Version == TombRaider_4) {
      /* XXX -- TR4: write NumZZ (bitu32) and ZZ[NumZZ] (16) */
      }

   /* write boxes (?) */
   TR2fwrite1_bit32(&Level->NumBoxes, fp);
   if (Level->NumBoxes > 0) {
      TR2fwrite_box(Level->Boxes, Level->NumBoxes, fp);
      }
   PRINTF(("wrote %ld Boxes\n", Level->NumBoxes));
   DUMPWR(fp, "Boxes");

   /* write overlaps (?) */
   TR2fwrite1_bit32(&Level->NumOverlaps, fp);
   if (Level->NumOverlaps > 0) {
      TR2fwrite_bit16(Level->Overlaps, Level->NumOverlaps, fp);
      }
   PRINTF(("wrote %ld Overlaps\n", Level->NumOverlaps));
   DUMPWR(fp, "Overlaps");

   /* write Zones (?) */
   if (Level->NumBoxes > 0) {
      TR2fwrite_bit16(Level->Zones, ((fp->Version==TombRaider_1)? 6: 10) * Level->NumBoxes, fp);
      }
   PRINTF(("wrote %ld Zones\n", ((fp->Version == TombRaider_1)? 6: 10) * Level->NumBoxes));
   DUMPWR(fp, "Zones");

   /* write animation textures (?) */
   TR2fwrite1_bit32(&Level->NumAnimatedTextures, fp);
   if (Level->NumAnimatedTextures > 0) {
      TR2fwrite_bit16(Level->AnimatedTextures, Level->NumAnimatedTextures, fp);
      }
   PRINTF(("wrote %ld AnimatedTextures\n", Level->NumAnimatedTextures));
   DUMPWR(fp, "AnimatedTextures");

   if (fp->Version == TombRaider_4) {
      /* XXX -- TR4: write "TEX" */
      bitu8 zzbuf[4] = {'T', 'E', 'X', '\0'};
      TR2fwrite_bitu8(zzbuf, 3, fp); // write "TEX"
      } 

   if (fp->Version >= TombRaider_3) {
      /* write object textures */
      TR2fwrite1_bitu32(&Level->NumObjectTextures, fp);
      if (Level->NumObjectTextures > 0) {
           TR2fwrite_object_texture(Level->ObjectTextures, Level->NumObjectTextures, fp);
         }
      PRINTF(("wrote %ld ObjectTextures\n", Level->NumObjectTextures));
      DUMPWR(fp, "ObjectTextures");
      }

   /* write items (?) */
   TR2fwrite1_bit32(&Level->NumItems, fp);
   if (Level->NumItems > 0) {
      TR2fwrite_item(Level->Items, Level->NumItems, fp);
      }
   PRINTF(("wrote %ld Items\n", Level->NumItems));
   DUMPWR(fp, "Items");

   /* write LightMaps */
   if (fp->Version != TombRaider_4) {
      TR2fwrite_bitu8(Level->LightMap, 32 * 256, fp);
      PRINTF(("wrote LightMap[%u]\n", 32 * 256));
      DUMPWR(fp, "LightMap");
      }

   if (fp->Version == TombRaider_1) {
      /* write the 8-bit palette */
      TR2fwrite_colour(Level->Palette8, dimof(Level->Palette8), fp);
      PRINTF(("wrote Palette8[%u]\n", dimof(Level->Palette8)));
      DUMPWR(fp, "Palette8");
      }

   /* write cinematic frames */
   if (fp->Version == TombRaider_4) {
      bitu32 NumCinematicFrames; // it's now bitu32
      NumCinematicFrames = (bitu32) Level->NumCinematicFrames;
      TR2fwrite1_bitu32(&NumCinematicFrames, fp);
      }
   else {
      TR2fwrite1_bitu16(&Level->NumCinematicFrames, fp);
      }
   if (Level->NumCinematicFrames > 0) {
      TR2fwrite_cinematic_frame(Level->CinematicFrames, Level->NumCinematicFrames, fp);
      }
   PRINTF(("wrote %d CinematicFrames\n", Level->NumCinematicFrames));
   DUMPWR(fp, "CinematicFrames");

   /* write demodata (?) */
   TR2fwrite1_bit16(&Level->NumDemoData, fp);
   if (Level->NumDemoData > 0) {
      TR2fwrite_bitu8(Level->DemoData, Level->NumDemoData, fp);
      }
   PRINTF(("wrote %d DemoData\n", Level->NumDemoData));
   DUMPWR(fp, "DemoData");

   /* write SoundMap */
   TR2fwrite_bit16(Level->SoundMap, (fp->Version==TombRaider_1)? 256: 370, fp);
   PRINTF(("wrote SoundMap[%d]\n", (fp->Version==TombRaider_1)? 256: 370));
   DUMPWR(fp, "SoundMap");

   /* write SoundDetails */
   TR2fwrite1_bit32(&Level->NumSoundDetails, fp);
   if (Level->NumSoundDetails > 0) {
      TR2fwrite_sound_details(Level->SoundDetails, Level->NumSoundDetails, fp);
      }
   PRINTF(("wrote %ld SoundDetails\n", Level->NumSoundDetails));
   DUMPWR(fp, "SoundDetails");

   if (fp->Version != TombRaider_1) {  // KLUDGE!
      /* write sampleindices */
      TR2fwrite1_bit32(&Level->NumSampleIndices, fp);
      if (Level->NumSampleIndices > 0) { 
         TR2fwrite_bit32(Level->SampleIndices, Level->NumSampleIndices, fp);
         }
      PRINTF(("wrote %ld SampleIndices\n", Level->NumSampleIndices));
      DUMPWR(fp, "SampleIndices");
      }

   return;
}

/*
 * The level writer.  Takes a level and filename (presumably SOMETHING.TR2),
 * writes the level to the file in TR2 format.
 *
 * Returns   0 if all went well
 *          -1 if there were any errors (writes to stderr in that case)
 */

int
WriteTR2level(TR2_Level *Level, char *FileName)
{
   FILE      *ffp;
   tr2_file  tf, *fp = &tf;

   /* try to open the file */
   if ((ffp = fopen(FileName, WRITEBINARY)) == NULL) {
      perror(FileName);
      return(-1);
      }
   TR2fmake(fp, Level->EngineVersion, ffp, NULL, 0);

   /* write the level */
   TR2fwrite_Level(Level, fp);

   /* close the file */
   fclose(ffp);

   return(0);
}

